"""
This module contains the main window of the application

luckyLUKS Copyright (c) 2014,2015,2022 Jasper van Hoorn (muzius@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <http://www.gnu.org/licenses/>
"""

import sys
import os.path

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QApplication, QWidget, QMainWindow, QDesktopWidget, QDialog,\
    QSystemTrayIcon, QMessageBox, QMenu, QAction, QLabel, QPushButton, QGridLayout, QStyle
from PyQt5.QtGui import QIcon

from luckyLUKS import utils, PROJECT_URL
from luckyLUKS.unlockUI import UnlockContainerDialog, UserInputError
from luckyLUKS.utilsUI import show_alert


class MainWindow(QMainWindow):

    """ A window that shows the current status of the encrypted container
        and a button to unlock/close it. Open containers
        leave an icon in the systray as a reminder to close them eventually.
    """

    def __init__(self, device_name=None, container_path=None, key_file=None, mount_point=None):
        """ Command line arguments checks are done here to be able to display a graphical dialog with error messages .
            If no arguments were supplied on the command line a setup dialog will be shown.
            All commands will be executed from a separate worker process with administrator privileges
            that gets initialized here.
            :param device_name: The device mapper name
            :type device_name: str/unicode or None
            :param container_path: The path of the container file
            :type container_path: str/unicode or None
            :param key_file: The path of an optional key file
            :type key_file: str/unicode or None
            :param mount_point: The path of an optional mount point
            :type mount_point: str/unicode or None
        """
        super().__init__()

        self.luks_device_name = device_name
        self.encrypted_container = container_path
        self.key_file = key_file
        self.mount_point = mount_point

        self.worker = None
        self.is_waiting_for_worker = False
        self.is_unlocked = False
        self.is_initialized = False
        self.has_tray = QSystemTrayIcon.isSystemTrayAvailable()

        # L10n: program name - translatable for startmenu titlebar etc
        self.setWindowTitle(_('luckyLUKS'))
        self.setWindowIcon(QIcon.fromTheme('dialog-password', QApplication.style().standardIcon(QStyle.SP_DriveHDIcon)))

        # check if cryptsetup and sudo are installed
        not_installed_msg = _('{program_name} executable not found!\n'
                              'Please install, eg for Debian/Ubuntu\n`apt-get install {program_name}`')
        if not utils.is_installed('cryptsetup'):
            show_alert(self, not_installed_msg.format(program_name='cryptsetup'), critical=True)
        if not utils.is_installed('sudo'):
            show_alert(self, not_installed_msg.format(program_name='sudo'), critical=True)
        # quick sanity checks before asking for passwd
        if os.getuid() == 0:
            show_alert(self,
                       _('Graphical programs should not be run as root!\nPlease call as normal user.'),
                       critical=True
                       )
        if self.encrypted_container and not os.path.exists(self.encrypted_container):
            show_alert(
                self,
                _('Container file not accessible\nor path does not exist:\n\n{file_path}')
                .format(file_path=self.encrypted_container),
                critical=True
            )

        # only either encrypted_container or luks_device_name supplied
        if bool(self.encrypted_container) != bool(self.luks_device_name):
            show_alert(self, _('Invalid arguments:\n'
                               'Please call without any arguments\n'
                               'or supply both container and name.\n\n'
                               '<b>{executable} -c CONTAINER -n NAME [-m MOUNTPOINT]</b>\n\n'
                               'CONTAINER = Path of the encrypted container file\n'
                               'NAME = A (unique) name to identify the unlocked container\n'
                               'Optional: MOUNTPOINT = where to mount the encrypted filesystem\n\n'
                               'If automatic mounting is configured on your system,\n'
                               'explicitly setting a mountpoint is not required\n\n'
                               'For more information, visit\n'
                               '<a href="{project_url}">{project_url}</a>'
                               ).format(executable=os.path.basename(sys.argv[0]),
                                        project_url=PROJECT_URL), critical=True)

        # spawn worker process with root privileges
        try:
            self.worker = utils.WorkerMonitor(self)
            # start communication thread
            self.worker.start()
        except utils.SudoException as se:
            show_alert(self, str(se), critical=True)
            return

        # if no arguments supplied, display dialog to gather this information
        if self.encrypted_container is None and self.luks_device_name is None:

            from luckyLUKS.setupUI import SetupDialog
            sd = SetupDialog(self)

            if sd.exec_() == QDialog.Accepted:
                self.luks_device_name = sd.get_luks_device_name()
                self.encrypted_container = sd.get_encrypted_container()
                self.mount_point = sd.get_mount_point()
                self.key_file = sd.get_keyfile()

                self.is_unlocked = True  # all checks in setup dialog -> skip initializing state
            else:
                # user closed dialog -> quit program
                # and check if a keyfile create thread has to be stopped
                # the worker process terminates itself when its parent dies
                if hasattr(sd, 'create_thread') and sd.create_thread is not None and sd.create_thread.isRunning():
                    sd.create_thread.terminate()
                QApplication.instance().quit()
                return

        # center window on desktop
        qr = self.frameGeometry()
        cp = QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())

        # widget content
        main_grid = QGridLayout()
        main_grid.setSpacing(10)
        icon = QLabel()
        icon.setPixmap(QIcon.fromTheme(
            'dialog-password',
            QApplication.style().standardIcon(QStyle.SP_DriveHDIcon)
        ).pixmap(32))
        main_grid.addWidget(icon, 0, 0)
        main_grid.addWidget(QLabel('<b>' + _('Handle encrypted container') + '</b>\n'), 0, 1, alignment=Qt.AlignCenter)

        main_grid.addWidget(QLabel(_('Name:')), 1, 0)
        main_grid.addWidget(QLabel('<b>{dev_name}</b>'.format(dev_name=self.luks_device_name)),
                            1, 1, alignment=Qt.AlignCenter)

        main_grid.addWidget(QLabel(_('File:')), 2, 0)
        main_grid.addWidget(QLabel(self.encrypted_container), 2, 1, alignment=Qt.AlignCenter)

        if self.key_file is not None:
            main_grid.addWidget(QLabel(_('Key:')), 3, 0)
            main_grid.addWidget(QLabel(self.key_file), 3, 1, alignment=Qt.AlignCenter)

        if self.mount_point is not None:
            main_grid.addWidget(QLabel(_('Mount:')), 4, 0)
            main_grid.addWidget(QLabel(self.mount_point), 4, 1, alignment=Qt.AlignCenter)

        main_grid.addWidget(QLabel(_('Status:')), 5, 0)
        self.label_status = QLabel('')
        main_grid.addWidget(self.label_status, 5, 1, alignment=Qt.AlignCenter)

        self.button_toggle_status = QPushButton('')
        self.button_toggle_status.setMinimumHeight(34)
        self.button_toggle_status.clicked.connect(self.toggle_container_status)
        main_grid.setRowMinimumHeight(6, 10)
        main_grid.addWidget(self.button_toggle_status, 7, 1)

        widget = QWidget()
        widget.setLayout(main_grid)
        widget.setContentsMargins(10, 10, 10, 10)
        self.setCentralWidget(widget)

        # tray popup menu
        if self.has_tray:
            tray_popup = QMenu(self)
            tray_popup.addAction(QIcon.fromTheme('dialog-password',
                                                 QApplication.style().standardIcon(QStyle.SP_DriveHDIcon)),
                                 self.luks_device_name).setEnabled(False)
            tray_popup.addSeparator()
            self.tray_toggle_action = QAction(QApplication.style().standardIcon(QStyle.SP_DesktopIcon), _('Hide'), self)
            self.tray_toggle_action.triggered.connect(self.toggle_main_window)
            tray_popup.addAction(self.tray_toggle_action)
            quit_action = QAction(QApplication.style().standardIcon(QStyle.SP_MessageBoxCritical), _('Quit'), self)
            quit_action.triggered.connect(self.tray_quit)
            tray_popup.addAction(quit_action)
            # systray
            self.tray = QSystemTrayIcon(self)
            self.tray.setIcon(QIcon.fromTheme(
                'dialog-password',
                QApplication.style().standardIcon(QStyle.SP_DriveHDIcon))
            )
            self.tray.setContextMenu(tray_popup)
            self.tray.activated.connect(self.toggle_main_window)
            self.tray.show()

        self.init_status()

    def refresh(self):
        """ Update widgets to reflect current container status. Adds systray icon if needed """
        if self.is_unlocked:
            self.label_status.setText(_('Container is {unlocked_green_bold}').format(
                unlocked_green_bold='<font color="#006400"><b>' + _('unlocked') + '</b></font>'))
            self.button_toggle_status.setText(_('Close Container'))
            if self.has_tray:
                self.tray.setToolTip(_('{device_name} is unlocked').format(device_name=self.luks_device_name))
        else:
            self.label_status.setText(_('Container is {closed_red_bold}').format(
                closed_red_bold='<font color="#b22222"><b>' + _('closed') + '</b></font>'))
            self.button_toggle_status.setText(_('Unlock Container'))
            if self.has_tray:
                self.tray.setToolTip(_('{device_name} is closed').format(device_name=self.luks_device_name))

        self.show()
        self.setFixedSize(self.sizeHint())

    def tray_quit(self):
        """ Triggered by clicking on `quit` in the systray popup: asks to close an unlocked container """
        if not self.is_unlocked:
            QApplication.instance().quit()
        elif not self.is_waiting_for_worker:
            self.show()
            self.confirm_close()

    def toggle_main_window(self, tray_icon_clicked):
        """ Triggered by clicking on the systray icon: show/hide main window """
        # don't activate on rightclick/contextmenu
        if not tray_icon_clicked or tray_icon_clicked == QSystemTrayIcon.Trigger:
            if self.isVisible():
                self.hide()
                self.tray_toggle_action.setText(_('Show'))
            else:
                self.show()
                self.tray_toggle_action.setText(_('Hide'))

    def closeEvent(self, event):
        """ Triggered by closing the window: If the container is unlocked,
            the program won't quit but remain in the systray. """
        if not self.is_waiting_for_worker:
            if self.is_unlocked:
                if self.has_tray:
                    self.hide()
                    self.tray_toggle_action.setText(_('Show'))
                else:
                    self.confirm_close()
                event.ignore()
            else:
                event.accept()

    def confirm_close(self):
        """ Inform about opened container and ask for confirmation to close & quit """
        message = _('<b>{device_name}</b> >> {container_path}\n'
                    'is currently <b>unlocked</b>,\n'
                    'Close Container now and quit?').format(device_name=self.luks_device_name,
                                                            container_path=self.encrypted_container)
        mb = QMessageBox(QMessageBox.Question, '', message, QMessageBox.Ok | QMessageBox.Cancel, self)
        mb.button(QMessageBox.Ok).setText(_('Quit'))
        if mb.exec_() == QMessageBox.Ok:
            self.do_close_container(shutdown=True)

    def customEvent(self, event):
        """ Receives response from worker and calls supplied callback function """
        event.callback(event.response)

    def toggle_container_status(self):
        """ Unlock or close container """
        if self.is_unlocked:
            self.do_close_container()
        else:
            try:
                UnlockContainerDialog(
                    self, self.worker, self.luks_device_name,
                    self.encrypted_container, self.key_file,
                    self.mount_point
                ).communicate()
                self.is_unlocked = True
            except UserInputError as uie:
                show_alert(self, str(uie))
                self.is_unlocked = False
            self.refresh()

    def do_close_container(self, shutdown=False):
        """ Send close command to worker and supply callbacks
            :param shutdown: Quit application after container successfully closed? (default=False)
            :type shutdown: bool
        """
        self.disable_ui(_('Closing Container ..'))
        self.worker.execute(command={'type': 'request',
                                     'msg': 'close',
                                     'device_name': self.luks_device_name,
                                     'container_path': self.encrypted_container
                                     },
                            success_callback=lambda msg: self.on_container_closed(msg, error=False, shutdown=shutdown),
                            error_callback=lambda msg: self.on_container_closed(msg, error=True, shutdown=shutdown))

    def on_container_closed(self, message, error, shutdown):
        """ Callback after worker closed container
            :param message: Contains an error description if error=True,
                            otherwise the current state of the container (unlocked/closed)
            :type message: str
            :param error: Error during closing of container
            :type error: bool
            :param shutdown: Quit application after container successfully closed?
            :type shutdown: bool
        """
        if error:
            show_alert(self, message)
        else:
            self.is_unlocked = False
        if not error and shutdown:  # automatic shutdown only if container successfully closed
            QApplication.instance().quit()
        else:
            self.enable_ui()

    def init_status(self):
        """ Request current status of container from worker if needed """
        if not self.is_unlocked:
            self.disable_ui(_('Initializing ..'))
            self.worker.execute(command={'type': 'request',
                                         'msg': 'status',
                                         'device_name': self.luks_device_name,
                                         'container_path': self.encrypted_container,
                                         'key_file': self.key_file,
                                         'mount_point': self.mount_point
                                         },
                                success_callback=self.on_initialized,
                                error_callback=lambda msg: self.on_initialized(msg, error=True))
        else:  # unlocked by setup-dialog -> just refresh UI
            self.enable_ui()
        self.is_initialized = True  # qt event loop can start now

    def on_initialized(self, message, error=False):
        """ Callback after worker send current state of container
            :param message: Contains an error description if error=True,
                            otherwise the current state of the container (unlocked/closed)
            :type message: str
            :param critical: Error during initialization (default=False)
            :type critical: bool
        """
        if error:
            show_alert(self, message, critical=True)
        else:
            self.is_unlocked = bool(message == 'unlocked')
            self.enable_ui()

    def enable_ui(self):
        """ Enable buttons and refresh state """
        self.refresh()
        self.is_waiting_for_worker = False
        self.button_toggle_status.setEnabled(True)

    def disable_ui(self, reason):
        """ Disable buttons and display waiting message
            :param reason: A waiting message that gets displayed
            :type reason: str/unicode
        """
        self.is_waiting_for_worker = True
        self.button_toggle_status.setText(reason)
        self.button_toggle_status.setEnabled(False)



"""
UI helper/classes for luckyLUKS

luckyLUKS Copyright (c) 2014,2015,2022 Jasper van Hoorn (muzius@gmail.com)
QExpander Copyright (c) 2012 Canonical Ltd.
modified, originally from https://launchpad.net/ubuntu-sso-client (GPL v3+)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <http://www.gnu.org/licenses/>
"""

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import QApplication, QMessageBox, QDialog, QVBoxLayout, QHBoxLayout,\
    QWidget, QDialogButtonBox, QLabel, QStyle, QStyleOption, QSizePolicy, QFrame
from PyQt5.QtGui import QPainter

from luckyLUKS import VERSION_STRING, PROJECT_URL


class HelpDialog(QDialog):

    """ Displays a help dialog that consists of
        a help icon and a header/title
        a main text
        an iniatially hidden secondary help text that can be expanded
        followed by a footer
    """

    def __init__(self, parent, header_text, basic_text, advanced_topics):
        """ Create a new instance
            :param parent: The parent window/dialog used to enable modal behaviour
            :type parent: :class:`PyQt5.QtGui.QWidget`
            :param header_text: Displayed in the top of the dialog next to the help icon
            :type header_text: str
            :param basic_text: Displayed in the middle of the help dialog
            :type basic_text: str
            :param advanced_topics: Displayed below the basic text, initially only header is shown and content hidden
            :type advanced_topics: Array of dicts with str head and text properties
        """
        super().__init__(parent, Qt.WindowCloseButtonHint | Qt.WindowTitleHint)
        self.setWindowTitle(_('Help'))
        layout = QVBoxLayout()
        layout.setContentsMargins(15, 5, 15, 5)
        layout.setSpacing(5)
        # icon and header
        header = QHBoxLayout()
        header.setSpacing(80)
        header.setAlignment(Qt.AlignLeft)
        icon = QLabel()
        icon.setPixmap(QApplication.style().standardIcon(QStyle.SP_DialogHelpButton).pixmap(48))
        header.addWidget(icon)
        header.addWidget(QLabel(header_text))
        layout.addLayout(header)
        # main help text
        basic_text = QLabel(basic_text)
        basic_text.setWordWrap(True)
        basic_text.setAlignment(Qt.AlignJustify)
        basic_text.setFixedWidth(470)  # Qt produces unreliable layout when using wordwrap and non-fixed width
        basic_text.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        layout.addWidget(basic_text)
        # advanced help
        advanced = QLabel('<b>' + _('Advanced Topics:') + '</b>')
        layout.addWidget(advanced)
        self._advanced_topics = []

        for topic in advanced_topics:
            head = QExpander(topic['head'], self, False)
            layout.addWidget(head)
            text = QLabel(topic['text'])
            text.setWordWrap(True)
            text.setAlignment(Qt.AlignJustify)
            text.setFixedWidth(470)
            text.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            layout.addWidget(text)
            head.addWidgets([text])
            self._advanced_topics += [head]
            head.clicked.connect(self._on_topic_clicked)

        # footer
        layout.addStretch()
        hl = QFrame()
        hl.setFrameShape(QFrame.HLine)
        hl.setFrameShadow(QFrame.Sunken)
        layout.addWidget(hl)

        footer = QLabel(_('luckyLUKS version {version}\n'
                          'For more information, visit\n'
                          '<a href="{project_url}">{project_url}</a>').format(version=VERSION_STRING,
                                                                              project_url=PROJECT_URL))
        footer.setContentsMargins(0, 10, 0, 10)
        layout.addWidget(footer)
        # button
        ok_button = QDialogButtonBox(QDialogButtonBox.Ok, parent=self)
        ok_button.accepted.connect(self.accept)
        layout.addWidget(ok_button)

        self.setLayout(layout)

    def _on_topic_clicked(self, clicked_topic):
        """An expandable topic was clicked. Closes previously opened topic if necessary"""
        if clicked_topic.is_expanded:
            for topic in self._advanced_topics:
                if not topic == clicked_topic:
                    topic.setExpanded(False)


def show_info(parent, message, title=''):
    """ Helper to show info message
        :param parent: The parent widget to be passed to the modal dialog
        :type parent: :class:`PyQt5.QtGui.QWidget`
        :param message: The message that gets displayed in a modal dialog
        :type message: str
        :param title: Displayed in the dialogs titlebar
        :type title: str
    """
    show_message(parent, message, title, QMessageBox.Information)


def show_alert(parent, message, critical=False):
    """ Helper to show error message
        :param parent: The parent widget to be passed to the modal dialog
        :type parent: :class:`PyQt5.QtGui.QWidget`
        :param message: The message that gets displayed in a modal dialog
        :type message: str
        :param critical: If critical, quit application (default=False)
        :type critical: bool
    """
    show_message(parent, message, _('Error'), QMessageBox.Critical if critical else QMessageBox.Warning)
    if critical:
        QApplication.instance().quit()


def show_message(parent, message, title, message_type):
    """ Generic helper to show message
        :param parent: The parent widget to be passed to the modal dialog
        :type parent: :class:`PyQt5.QtGui.QWidget`
        :param message: The message that gets displayed in a modal dialog
        :type message: str
        :param title: Displayed in the dialogs titlebar
        :type title: str
        :param message_type: Type of message box to be used
        :type message_type: :class:`QMessageBox.Icon`
    """
    if message != '':
        mb = QMessageBox(message_type, title, message, QMessageBox.Ok, parent)
        # make QMessageBox better adaptable to long messages (eg stacktraces)
        mb.findChildren(QLabel)[1].setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)
        mb.exec_()


class QExpander(QWidget):

    """A Qt implementation similar to GtkExpander."""

    clicked = pyqtSignal(QWidget)

    def __init__(self, label, parent, expanded=False):
        """Create a new instance."""
        super().__init__(parent)
        self.parent = parent
        self.label = QExpanderLabel(label, self)
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)
        self.layout.addWidget(self.label)
        self._widgets = []
        self._expanded = False
        self.label.clicked.connect(self._on_label_clicked)
        self.setExpanded(expanded)

    def _on_label_clicked(self):
        """The expander widget was clicked."""
        self._expanded = not self._expanded
        self.setExpanded(self._expanded)
        self.clicked.emit(self)

    def addWidgets(self, widgets):
        """Add widgets to the expander.
        """
        self._widgets += widgets
        self.setExpanded(self._expanded)

    def is_expanded(self):
        """Return if widget is expanded."""
        return self._expanded

    def setExpanded(self, is_expanded):
        """Expand the widget or not."""
        self._expanded = is_expanded
        if self._expanded:
            self.label.arrow.direction = QArrow.DOWN
        else:
            self.label.arrow.direction = QArrow.RIGHT
        for widget in self._widgets:
            widget.setVisible(self._expanded)
        self.parent.adjustSize()


class QExpanderLabel(QWidget):

    """Widget used to show/modify the label of a QExpander."""

    clicked = pyqtSignal()

    def __init__(self, label, parent):
        """Create a new instance."""
        super().__init__(parent)
        self.arrow = QArrow(QArrow.RIGHT)
        self.label = QLabel(label)
        layout = QHBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.arrow)
        layout.addWidget(self.label)

    def mousePressEvent(self, event):
        """Mouse clicked."""
        if self.arrow.direction == QArrow.DOWN:
            self.arrow.direction = QArrow.RIGHT
        else:
            self.arrow.direction = QArrow.DOWN
        self.clicked.emit()


class QArrow(QWidget):

    """Custom widget, arrow image that can be pointed in 4 different directions"""

    UP = 0
    DOWN = 1
    LEFT = 2
    RIGHT = 3

    def __init__(self, direction, parent=None):
        """Create a new instance."""
        super().__init__(parent)
        self._set_direction(direction)
        self.setFixedWidth(10)

    def paintEvent(self, event):
        """Paint the widget."""
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        if self._direction == QArrow.UP:
            primitive = QStyle.PE_IndicatorArrowUp
        elif self._direction == QArrow.DOWN:
            primitive = QStyle.PE_IndicatorArrowDown
        elif self._direction == QArrow.LEFT:
            primitive = QStyle.PE_IndicatorArrowLeft
        else:
            primitive = QStyle.PE_IndicatorArrowRight
        painter.setViewTransformEnabled(True)
        self.style().drawPrimitive(primitive, opt, painter, self)

    def _get_direction(self):
        """Return the direction used."""
        return self._direction

    def _set_direction(self, direction):
        """Set the direction."""
        if direction not in (QArrow.UP, QArrow.DOWN,
                             QArrow.LEFT, QArrow.RIGHT):
            raise ValueError('Wrong arrow direction.')
        self._direction = direction
        self.repaint()

    direction = property(_get_direction, _set_direction)



"""
If the main program gets called without arguments, this GUI will be shown first.

luckyLUKS Copyright (c) 2014,2015,2022 Jasper van Hoorn (muzius@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <http://www.gnu.org/licenses/>
"""

import os
import sys
import codecs
import random
import string
import subprocess
import glob
from shutil import move

from PyQt5.QtCore import QTimer, Qt
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QTabWidget, QDialogButtonBox, QGridLayout, QLabel, QStackedWidget,\
    QMessageBox, QLineEdit, QPushButton, QSpinBox, QComboBox, QCheckBox, QFileDialog, QWidget, QStyle, QApplication,\
    QProgressBar, QLayout

from luckyLUKS.unlockUI import FormatContainerDialog, UnlockContainerDialog, UserInputError
from luckyLUKS.utilsUI import QExpander, HelpDialog, show_info, show_alert
from luckyLUKS.utils import is_installed


class SetupDialog(QDialog):

    """ This dialog consists of two parts/tabs: The first one is supposed to help choosing
        container, device name and mount point to unlock an existing container.
        The second tab assists in creating a new encrypted LUKS container.
    """

    def __init__(self, parent):
        """ :param parent: The parent window/dialog used to enable modal behaviour
            :type parent: :class:`PyQt5.QtGui.QWidget`
        """
        super().__init__(parent, Qt.WindowCloseButtonHint | Qt.WindowTitleHint)
        self.setWindowTitle(_('luckyLUKS'))

        self.worker = parent.worker
        self.is_busy = False

        # build ui
        self.layout = QVBoxLayout()
        self.layout.setSizeConstraint(QLayout.SetFixedSize)
        style = QApplication.style()
        # set up stacked layout: initially only the main tab pane (unlock/create)
        self.main_pane = QStackedWidget()
        self.tab_pane = QTabWidget()
        self.main_pane.addWidget(self.tab_pane)
        self.layout.addWidget(self.main_pane)
        self.create_pane = None  # init in on_create_container()

        # Unlock Tab
        unlock_grid = QGridLayout()
        unlock_grid.setColumnMinimumWidth(1, 220)
        uheader = QLabel(_('<b>Unlock an encrypted container</b>\n') +
                         _('Please select container file and name'))
        uheader.setContentsMargins(0, 10, 0, 10)
        unlock_grid.addWidget(uheader, 0, 0, 1, 3, Qt.AlignCenter)

        label = QLabel(_('container file'))
        label.setIndent(5)
        unlock_grid.addWidget(label, 1, 0)
        self.unlock_container_file = QLineEdit()
        unlock_grid.addWidget(self.unlock_container_file, 1, 1)
        button_choose_file = QPushButton(style.standardIcon(QStyle.SP_DialogOpenButton), '', self)
        button_choose_file.setToolTip(_('choose file'))
        unlock_grid.addWidget(button_choose_file, 1, 2)
        button_choose_file.clicked.connect(self.on_select_container_clicked)

        label = QLabel(_('device name'))
        label.setIndent(5)
        unlock_grid.addWidget(label, 2, 0)
        self.unlock_device_name = QLineEdit()
        unlock_grid.addWidget(self.unlock_device_name, 2, 1)
        # advanced settings
        a_settings = QExpander(_('Advanced'), self, False)
        unlock_grid.addWidget(a_settings, 3, 0, 1, 3)

        label = QLabel(_('key file'))
        label.setIndent(5)
        unlock_grid.addWidget(label, 4, 0)
        self.unlock_keyfile = QLineEdit()
        unlock_grid.addWidget(self.unlock_keyfile, 4, 1)
        button_choose_uKeyfile = QPushButton(style.standardIcon(QStyle.SP_DialogOpenButton), '')
        button_choose_uKeyfile.setToolTip(_('choose keyfile'))
        unlock_grid.addWidget(button_choose_uKeyfile, 4, 2)
        button_choose_uKeyfile.clicked.connect(lambda: self.on_select_keyfile_clicked('Unlock'))
        a_settings.addWidgets([unlock_grid.itemAtPosition(4, column).widget() for column in range(0, 3)])

        label = QLabel(_('mount point'))
        label.setIndent(5)
        unlock_grid.addWidget(label, 5, 0)
        self.unlock_mountpoint = QLineEdit()
        unlock_grid.addWidget(self.unlock_mountpoint, 5, 1)
        button_choose_mountpoint = QPushButton(style.standardIcon(QStyle.SP_DialogOpenButton), '')
        button_choose_mountpoint.setToolTip(_('choose folder'))
        unlock_grid.addWidget(button_choose_mountpoint, 5, 2)
        button_choose_mountpoint.clicked.connect(self.on_select_mountpoint_clicked)
        a_settings.addWidgets([unlock_grid.itemAtPosition(5, column).widget() for column in range(0, 3)])

        unlock_grid.setRowStretch(6, 1)
        unlock_grid.setRowMinimumHeight(6, 10)
        button_help_unlock = QPushButton(style.standardIcon(QStyle.SP_DialogHelpButton), _('Help'))
        button_help_unlock.clicked.connect(self.show_help_unlock)
        unlock_grid.addWidget(button_help_unlock, 7, 2)

        unlock_tab = QWidget()
        unlock_tab.setLayout(unlock_grid)
        self.tab_pane.addTab(unlock_tab, _('Unlock Container'))

        # Create Tab
        create_grid = QGridLayout()
        cheader = QLabel(_('<b>Create a new encrypted container</b>\n') +
                         _('Please choose container file, name and size'))
        cheader.setContentsMargins(0, 10, 0, 10)
        create_grid.addWidget(cheader, 0, 0, 1, 3, Qt.AlignCenter)

        label = QLabel(_('container file'))
        label.setIndent(5)
        create_grid.addWidget(label, 1, 0)
        self.create_container_file = QLineEdit()
        create_grid.addWidget(self.create_container_file, 1, 1)
        button_choose_file = QPushButton(style.standardIcon(QStyle.SP_DialogOpenButton), '')
        button_choose_file.setToolTip(_('set file'))
        create_grid.addWidget(button_choose_file, 1, 2)
        button_choose_file.clicked.connect(self.on_save_container_clicked)

        label = QLabel(_('device name'))
        label.setIndent(5)
        create_grid.addWidget(label, 2, 0)
        self.create_device_name = QLineEdit()
        create_grid.addWidget(self.create_device_name, 2, 1)

        label = QLabel(_('container size'))
        label.setIndent(5)
        create_grid.addWidget(label, 3, 0)
        self.create_container_size = QSpinBox()
        self.create_container_size.setRange(1, 1000000000)
        self.create_container_size.setValue(1)
        create_grid.addWidget(self.create_container_size, 3, 1)

        self.create_size_unit = QComboBox()
        self.create_size_unit.addItems(['MB', 'GB'])
        self.create_size_unit.setCurrentIndex(1)
        create_grid.addWidget(self.create_size_unit, 3, 2)

        self.create_quickformat = QCheckBox(_('Quickformat'))
        create_grid.addWidget(self.create_quickformat, 4, 1)

        # advanced settings
        a_settings = QExpander(_('Advanced'), self, False)
        create_grid.addWidget(a_settings, 5, 0, 1, 3)

        label = QLabel(_('key file'))
        label.setIndent(5)
        create_grid.addWidget(label, 6, 0)
        self.create_keyfile = QLineEdit()
        create_grid.addWidget(self.create_keyfile, 6, 1)
        button_choose_cKeyfile = QPushButton(style.standardIcon(QStyle.SP_DialogOpenButton), '')
        button_choose_cKeyfile.setToolTip(_('choose keyfile'))
        create_grid.addWidget(button_choose_cKeyfile, 6, 2)
        button_choose_cKeyfile.clicked.connect(lambda: self.on_select_keyfile_clicked('Create'))
        a_settings.addWidgets([create_grid.itemAtPosition(6, column).widget() for column in range(0, 3)])

        button_create_keyfile = QPushButton(_('Create key file'))
        button_create_keyfile.clicked.connect(self.on_create_keyfile)
        create_grid.addWidget(button_create_keyfile, 7, 1)
        a_settings.addWidgets([button_create_keyfile])

        label = QLabel(_('format'))
        label.setIndent(5)
        create_grid.addWidget(label, 8, 0)
        self.create_encryption_format = QComboBox()
        self.create_encryption_format.addItem('LUKS')
        self.create_encryption_format.addItem('TrueCrypt')
        if not is_installed('tcplay'):
            self.create_encryption_format.setEnabled(False)
        self.create_encryption_format.setCurrentIndex(0)
        create_grid.addWidget(self.create_encryption_format, 8, 1)
        a_settings.addWidgets([create_grid.itemAtPosition(8, column).widget() for column in range(0, 2)])

        label = QLabel(_('filesystem'))
        label.setIndent(5)
        create_grid.addWidget(label, 9, 0)
        filesystems = ['ext4', 'ext2', 'ntfs']
        self.create_filesystem_type = QComboBox()
        for filesystem in filesystems:
            if is_installed('mkfs.' + filesystem):
                self.create_filesystem_type.addItem(filesystem)
        self.create_filesystem_type.setCurrentIndex(0)
        create_grid.addWidget(self.create_filesystem_type, 9, 1)
        a_settings.addWidgets([create_grid.itemAtPosition(9, column).widget() for column in range(0, 2)])

        create_grid.setRowStretch(10, 1)
        create_grid.setRowMinimumHeight(10, 10)
        button_help_create = QPushButton(style.standardIcon(QStyle.SP_DialogHelpButton), _('Help'))
        button_help_create.clicked.connect(self.show_help_create)
        create_grid.addWidget(button_help_create, 11, 2)

        create_tab = QWidget()
        create_tab.setLayout(create_grid)
        self.tab_pane.addTab(create_tab, _('Create New Container'))
        self.tab_pane.currentChanged.connect(self.on_switchpage_event)

        # OK and Cancel buttons
        self.buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, parent=self)
        self.buttons.button(QDialogButtonBox.Ok).setText(_('Unlock'))
        self.buttons.accepted.connect(self.on_accepted)
        self.buttons.rejected.connect(self.reject)
        self.layout.addWidget(self.buttons)

        # ui built, add to widget
        self.setLayout(self.layout)

        # var-def for create-subdialog
        self.create_progressbars = []
        self.create_thread = None
        self.create_status_grid = None
        self.create_timer = None

    def on_create_container(self):
        """ Triggered by clicking create.
            Hides the unlock/create pane and switches to a status pane
            where the progress in creating the new container can be followed step by step.
            This shows the header and the first step:
            Initializing the container file with random data
        """
        self.init_create_pane()

        header = QLabel(_('<b>Creating new container</b>\n') +
                        _('patience .. this might take a while'))
        header.setContentsMargins(0, 10, 0, 10)
        self.create_status_grid.addWidget(header, 0, 0, 1, 3, Qt.AlignCenter)

        self.create_status_grid.addWidget(QLabel('<b>' + _('Step') + ' 1/3</b>'), 1, 0)
        self.create_status_grid.addWidget(QLabel(_('Initializing Container File')), 1, 1)
        self.create_progressbars.append(QProgressBar())
        self.create_progressbars[0].setRange(0, 100)
        self.create_status_grid.addWidget(self.create_progressbars[0], 2, 0, 1, 3)
        self.create_status_grid.setRowStretch(7, 1)  # top align
        # add to stack widget and switch display
        self.main_pane.addWidget(self.create_pane)
        self.main_pane.setCurrentIndex(1)

        # calculate designated container size for worker and progress indicator
        size = self.create_container_size.value()
        size = size * (1024 * 1024 * 1024 if self.create_size_unit.currentIndex() == 1 else 1024 * 1024)  # GB vs MB
        location = self.create_container_file.text().strip()
        if not os.path.dirname(location):
            location = os.path.join(os.path.expanduser('~'), location)
            self.create_container_file.setText(location)
        keyfile = self.create_keyfile.text().strip() if self.create_keyfile.text().strip() != '' else None
        # start timer for progressbar updates during container creation
        self.create_timer.timeout.connect(lambda: self.display_progress_percent(location, size))
        self.create_timer.start(500)

        self.worker.execute(command={'type': 'request',
                                     'msg': 'create',
                                     'device_name': self.create_device_name.text().strip(),
                                     'container_path': location,
                                     'container_size': size,
                                     'quickformat': self.create_quickformat.isChecked(),
                                     'key_file': keyfile,
                                     'filesystem_type': str(self.create_filesystem_type.currentText()),
                                     'encryption_format': str(self.create_encryption_format.currentText()),
                                     },
                            success_callback=self.on_luksFormat_prompt,
                            error_callback=lambda msg: self.display_create_failed(msg, stop_timer=True))

    def on_luksFormat_prompt(self, msg):
        """ Triggered after the container file is created on disk
            Shows information about the next step and asks the user
            for the passphrase to be used with the new container
        """
        self.set_progress_done(self.create_timer, self.create_progressbars[0])

        self.create_status_grid.addWidget(QLabel('<b>' + _('Step') + ' 2/3</b>'), 3, 0)
        self.create_status_grid.addWidget(QLabel(_('Initializing Encryption')), 3, 1)
        self.create_progressbars.append(QProgressBar())
        self.create_progressbars[1].setRange(0, 0)
        self.create_status_grid.addWidget(self.create_progressbars[1], 4, 0, 1, 3)

        if msg == 'getPassword':
            try:
                self.worker.execute(command={'type': 'response',
                                             'msg': FormatContainerDialog(self).get_password()
                                             },
                                    success_callback=self.on_creating_filesystem,
                                    error_callback=self.display_create_failed)
            except UserInputError:  # user cancelled dlg
                self.worker.execute({'type': 'abort', 'msg': ''}, None, None)  # notify worker process
                self.display_create_failed(_('Initialize container aborted'))
        else:  # using keyfile
            self.worker.execute(command={'type': 'response', 'msg': ''},
                                success_callback=self.on_creating_filesystem,
                                error_callback=self.display_create_failed)

    def on_creating_filesystem(self, msg):
        """ Triggered after LUKS encryption got initialized.
            Shows information about the last step
        """
        self.set_progress_done(progressbar=self.create_progressbars[1])

        self.create_status_grid.addWidget(QLabel('<b>' + _('Step') + ' 3/3</b>'), 5, 0)
        self.create_status_grid.addWidget(QLabel(_('Initializing Filesystem')), 5, 1)
        self.create_progressbars.append(QProgressBar())
        self.create_progressbars[2].setRange(0, 0)
        self.create_status_grid.addWidget(self.create_progressbars[2], 6, 0, 1, 3)

        self.worker.execute(command={'type': 'response', 'msg': ''},
                            success_callback=self.display_create_success,
                            error_callback=self.display_create_failed)

    def display_create_success(self, msg):
        """ Triggered after successful creation of a new container """
        self.set_progress_done(progressbar=self.create_progressbars[2])
        # copy values of newly created container to unlock dlg und reset create values
        self.unlock_container_file.setText(self.create_container_file.text())
        self.unlock_device_name.setText(self.create_device_name.text())
        self.unlock_keyfile.setText(self.create_keyfile.text())
        show_info(self, _('<b>{device_name}\nsuccessfully created!</b>\nClick on unlock to use the new container')
                  .format(device_name=self.create_device_name.text().strip()), _('Success'))
        # reset create ui and switch to unlock tab
        self.create_container_file.setText('')
        self.create_device_name.setText('')
        self.create_container_size.setValue(1)
        self.create_size_unit.setCurrentIndex(1)
        self.create_keyfile.setText('')
        self.create_encryption_format.setCurrentIndex(0)
        self.create_filesystem_type.setCurrentIndex(0)
        self.display_create_done()
        self.tab_pane.setCurrentIndex(0)

    def display_create_failed(self, errormessage, stop_timer=False):
        """ Triggered when an error happend during the create process
            :param errormessage: errormessage to be shown
            :type errormessage: str
            :param stop_timer: stop a progress indicator?
            :type stop_timer: bool
        """
        if stop_timer:
            self.set_progress_done(self.create_timer)
        show_alert(self, errormessage)
        self.display_create_done()

    def display_create_done(self):
        """ Helper to hide the create process informations and show the unlock/create pane """
        self.is_busy = False
        self.main_pane.setCurrentIndex(0)
        self.buttons.setEnabled(True)

    def set_progress_done(self, timeout=None, progressbar=None):
        """ Helper to end stop the progress indicator
            :param timeout: Timer to stop
            :type timeout: QTimer or None
            :param progressbar: progressbar widget to set to 'Done'
            :type progressbar: :class:`QProgressBar` or None
        """
        if timeout is not None:
            timeout.stop()
        if progressbar is not None:
            if not progressbar.maximum():
                progressbar.setRange(0, 100)
            progressbar.setValue(100)
            progressbar.setFormat(_('Done'))

    def on_accepted(self):
        """ Event handler for response:
            Start unlock or create action
        """
        try:
            if self.tab_pane.currentIndex() == 1:

                self.on_create_container()

            else:
                UnlockContainerDialog(self,
                                      self.worker,
                                      self.get_luks_device_name(),
                                      self.get_encrypted_container(),
                                      self.get_keyfile(),
                                      self.get_mount_point()
                                      ).communicate()  # blocks

                # optionally create startmenu entry
                self.show_create_startmenu_entry()
                # all good, now switch to main window
                self.accept()

        except UserInputError as error:
            show_alert(self, str(error))

    def init_create_pane(self):
        """ Helper that initializes the ui for the progress indicators shown while creating containers or keyfiles """
        self.is_busy = True
        self.create_progressbars = []
        self.create_timer = QTimer(self)

        self.buttons.setEnabled(False)

        if self.main_pane.count() > 1:  # remove previous create display if any
            self.main_pane.removeWidget(self.create_pane)

        # built ui
        self.create_pane = QWidget()
        self.create_status_grid = QGridLayout()
        self.create_pane.setLayout(self.create_status_grid)
        self.create_status_grid.setVerticalSpacing(5)

    def on_create_keyfile(self):
        """ Triggered by clicking the `create key file` button below the key file text field (create)
            Asks for key file location if not already provided, creates the progress ui and starts a create-thread
        """
        if self.create_keyfile.text() == '':
            key_file = self.on_save_file(_('new_keyfile.bin')).strip()
        else:
            key_file = self.create_keyfile.text().strip()

        if not os.path.dirname(key_file):
            key_file = os.path.join(os.path.expanduser('~'), key_file)

        self.init_create_pane()

        header = QLabel(_('<b>Creating key file</b>'))
        self.create_status_grid.addWidget(header, 1, 0, 1, 3, Qt.AlignCenter)
        header.setContentsMargins(0, 30, 0, 10)

        self.create_progressbars.append(QProgressBar())
        self.create_progressbars[0].setRange(0, 100)
        self.create_status_grid.addWidget(self.create_progressbars[0], 2, 0, 1, 3)
        info = QLabel(_('This might take a while. Since computers are deterministic machines\n'
                        'it is quite a challenge to generate real random data for the key.\n'
                        '\n'
                        'You can speed up the process by typing, moving the mouse\n'
                        'and generally use the computer while the key gets generated.'))
        info.setContentsMargins(0, 10, 0, 10)
        self.create_status_grid.addWidget(info, 3, 0, 1, 3, Qt.AlignCenter)

        self.create_status_grid.setRowStretch(4, 2)  # vertical align
        # add to stack widget and switch display
        self.main_pane.addWidget(self.create_pane)
        self.main_pane.setCurrentIndex(1)

        # start timer for progressbar updates during keyfile creation
        self.create_timer.timeout.connect(lambda: self.display_progress_percent(key_file, 1024))
        self.create_timer.start(500)

        # run QThread with keyfile creation
        from luckyLUKS.utils import KeyfileCreator

        self.create_thread = KeyfileCreator(self, key_file)
        self.create_thread.start()

    def on_keyfile_created(self, key_file_path):
        """ Triggered when key file creation was successful. Restores the normal setup ui """
        self.set_progress_done(self.create_timer, progressbar=self.create_progressbars[0])
        show_info(self, _('<b>{key_file}\nsuccessfully created!</b>\n'
                          'You can use this key file now,\n'
                          'to create a new container.').format(key_file=key_file_path), _('Success'))
        self.display_create_done()
        self.create_keyfile.setText(key_file_path)

    def show_create_startmenu_entry(self):
        """ Shown after successfull unlock with setup dialog -> ask for shortcut creation """
        message = (_('<b>Successfully unlocked!</b>\n\n'
                     'Do you want to create\n'
                     'a startup menu entry for <b>{device_name}</b>?\n\n'
                     '-> Your password will NOT be saved!\n'
                     '   This just creates a shortcut,\n'
                     '   to the unlock container dialog.\n').format(
            device_name=self.get_luks_device_name())
        )
        mb = QMessageBox(QMessageBox.Question, '', message, QMessageBox.Ok | QMessageBox.Cancel, self)
        mb.button(QMessageBox.Ok).setText(_('Create shortcut'))
        mb.button(QMessageBox.Cancel).setText(_('No, thanks'))
        if mb.exec_() == QMessageBox.Ok:
            self.create_startmenu_entry()

    def create_startmenu_entry(self):
        """ Creates a startmenu entry that lets the user skip the setup dialog and go directly to the main UI
            Includes a workaround for the safety net some desktop environments create around the startupmenu
        """
        # command to be saved in shortcut: calling the script with the arguments entered in the dialog
        # put all arguments in single quotes and escape those in the strings (shell escape ' -> '\'')
        cmd = os.path.abspath(sys.argv[0])
        cmd += " -c '" + self.get_encrypted_container().replace("'", "'\\'''") + "'"
        cmd += " -n '" + self.get_luks_device_name().replace("'", "'\\'''") + "'"
        if self.get_mount_point() is not None:
            cmd += " -m '" + self.get_mount_point().replace("'", "'\\'''") + "'"
        if self.get_keyfile() is not None:
            cmd += " -k '" + self.get_keyfile().replace("'", "'\\'''") + "'"

        # create .desktop-file
        # xdg-desktop-menu has problems does not like some special chars
        filename = _('luckyLUKS') + '-' + ''.join(i for i in self.get_luks_device_name() if i not in ' \\/:*?<>|')
        if is_installed('xdg-desktop-menu'):  # create in tmp and add freedesktop menu entry
            # some desktop menus dont delete the .desktop files if a user removes
            # items from the menu but keep track of those files instead and do not readded later
            # -> the random part of the filename works around this behaviour
            rand4 = ''.join(random.choice(string.ascii_letters) for i in range(4))
            desktop_file_path = filename + '-' + rand4 + '.desktop'
            desktop_file_path = os.path.join('/tmp', desktop_file_path)
        else:  # or create in users home dir
            desktop_file_path = os.path.join(os.path.expanduser('~'), filename + '.desktop')

        desktop_file = codecs.open(desktop_file_path, 'w', 'utf-8')

        entry_name = _('Unlock {device_name}').format(device_name=self.get_luks_device_name())
        desktop_file.write("[Desktop Entry]\n")
        desktop_file.write("Name=" + entry_name + "\n")
        desktop_file.write("Comment=" + self.get_luks_device_name() + " " + _('Encrypted Container Tool') + "\n")
        desktop_file.write("GenericName=" + _('Encrypted Container') + "\n")
        desktop_file.write("Categories=Utility;\n")
        desktop_file.write("Exec=" + cmd + "\n")
        desktop_file.write("Icon=dialog-password\n")
        desktop_file.write("NoDisplay=false\n")
        desktop_file.write("StartupNotify=false\n")
        desktop_file.write("Terminal=0\n")
        desktop_file.write("TerminalOptions=\n")
        desktop_file.write("Type=Application\n\n")
        desktop_file.close()

        os.chmod(desktop_file_path, 0o700)  # some distros need the xbit to trust the desktop file

        if is_installed('xdg-desktop-menu'):
            # safest way to ensure updates: explicit uninstall
            # followed by installing a new desktop file with different random part
            desktopfiles = os.path.expanduser('~') + '/.local/share/applications/' + filename + '-*.desktop'
            for desktopfile in glob.glob(desktopfiles):
                with open(os.devnull) as DEVNULL:
                    subprocess.call(
                        ['xdg-desktop-menu', 'uninstall', desktopfile],
                        stdout=DEVNULL, stderr=subprocess.STDOUT)
            try:
                subprocess.check_output(
                    ['xdg-desktop-menu', 'install', '--novendor', desktop_file_path],
                    stderr=subprocess.STDOUT, universal_newlines=True)
                os.remove(desktop_file_path)  # remove from tmp
                show_info(self, _('<b>` {name} `</b>\nadded to start menu').format(name=entry_name), _('Success'))
            except subprocess.CalledProcessError as cpe:
                home_dir_path = os.path.join(os.path.expanduser('~'), os.path.basename(desktop_file_path))
                # move to homedir instead
                move(desktop_file_path, home_dir_path)
                show_alert(self, cpe.output)
                show_info(
                    self,
                    _('Adding to start menu not possible,\n'
                      'please place your shortcut manually.\n\nDesktop file saved to\n{location}')
                    .format(location=home_dir_path)
                )
        else:
            show_info(
                self,
                _('Adding to start menu not possible,\n'
                  'please place your shortcut manually.\n\nDesktop file saved to\n{location}')
                .format(location=desktop_file_path)
            )

    def reject(self):
        """ Event handler cancel: Ask for confirmation while creating container """
        if self.confirm_close():
            super().reject()

    def closeEvent(self, event):
        """ Event handler close: ask for confirmation while creating container """
        if not self.confirm_close():
            event.ignore()

    def confirm_close(self):
        """ Displays a confirmation dialog if currently busy creating container or keyfile
            :returns: The users decision or True if no create process running
            :rtype: bool
        """
        if self.is_busy:
            message = _('Currently processing your request!\nDo you really want to quit?')
            mb = QMessageBox(QMessageBox.Question, '', message, QMessageBox.Ok | QMessageBox.Cancel, self)
            mb.button(QMessageBox.Ok).setText(_('Quit'))
            return mb.exec_() == QMessageBox.Ok
        return True

    def on_switchpage_event(self, index):
        """ Event handler for tab switch: change text on OK button (Unlock/Create) """
        new_ok_label = _('Unlock')
        if index == 1:  # create
            if self.create_filesystem_type.currentText() == '':
                show_alert(self, _('No tools to format the filesystem found\n'
                                   'Please install, eg for Debian/Ubuntu\n'
                                   '`apt-get install e2fslibs ntfs-3g`'))
            new_ok_label = _('Create')
        self.buttons.button(QDialogButtonBox.Ok).setText(new_ok_label)

    def on_select_container_clicked(self):
        """ Triggered by clicking the select button next to container file (unlock) """
        file_path = QFileDialog.getOpenFileName(self, _('Please choose a container file'), os.getenv("HOME"))
        if isinstance(file_path, tuple):  # qt5 returns tuple path/selected filter
            file_path = file_path[0]
        self.unlock_container_file.setText(file_path)
        self.buttons.button(QDialogButtonBox.Ok).setText(_('Unlock'))

    def on_select_mountpoint_clicked(self):
        """ Triggered by clicking the select button next to mount point """
        self.unlock_mountpoint.setText(
            QFileDialog.getExistingDirectory(self, _('Please choose a folder as mountpoint'), os.getenv("HOME"))
        )
        self.buttons.button(QDialogButtonBox.Ok).setText(_('Unlock'))

    def on_select_keyfile_clicked(self, tab):
        """ Triggered by clicking the select button next to key file (both unlock and create tab) """
        file_path = QFileDialog.getOpenFileName(self, _('Please choose a key file'), os.getenv("HOME"))
        if isinstance(file_path, tuple):  # qt5 returns tuple path/selected filter
            file_path = file_path[0]
        if tab == 'Unlock':
            self.unlock_keyfile.setText(file_path)
        elif tab == 'Create':
            self.create_keyfile.setText(file_path)
        self.buttons.button(QDialogButtonBox.Ok).setText(_(tab))

    def on_save_container_clicked(self):
        """ Triggered by clicking the select button next to container file (create)
            Uses a file dialog to set the path of the container file to be created
        """
        self.create_container_file.setText(self.on_save_file(_('new_container.bin')))

    def on_save_file(self, default_filename):
        """ Opens a native file dialog and returns the chosen path of the file to be saved
            The dialog does not allow overwriting existing files - to get this behaviour
            while using native dialogs the QFileDialog has to be reopened on overwrite.
            A bit weird but enables visual consistency with the other file choose dialogs
            :param default_filename: The default filename to be used in the Qt file dialog
            :type default_filename: str
            :returns: The designated key file path
            :rtype: str
        """
        def_path = os.path.join(os.getenv("HOME"), default_filename)

        while True:
            save_path = QFileDialog.getSaveFileName(self,
                                                    _('Please create a new file'),
                                                    def_path,
                                                    options=QFileDialog.DontConfirmOverwrite)

            save_path = save_path[0].strip() if isinstance(save_path, tuple) else save_path.strip()
            self.buttons.button(QDialogButtonBox.Ok).setText(_('Create'))  # qt keeps changing this..

            if os.path.exists(save_path):
                show_alert(self, _('File already exists:\n{filename}\n\n'
                                   '<b>Please create a new file!</b>').format(filename=save_path))
                def_path = os.path.join(os.path.basename(save_path), default_filename)
            else:
                return save_path

    def display_progress_percent(self, location, size):
        """ Update value on the container creation progress bar
            :param location: The path of the container file currently being created
            :type location: str
            :param size: The final size the new container in bytes
            :type size: int
        """
        try:
            new_value = int(os.path.getsize(location) / size * 100)
        except (ZeroDivisionError, OSError, TypeError):
            new_value = 0
        self.create_progressbars[0].setValue(new_value)

    def get_encrypted_container(self):
        """ Getter for QLineEdit text
            :returns: The container file path
            :rtype: str
        """
        return self.unlock_container_file.text().strip()

    def get_luks_device_name(self):
        """ Getter for QLineEdit text
            :returns: The device name
            :rtype: str
        """
        return self.unlock_device_name.text().strip()

    def get_keyfile(self):
        """ Getter for QLineEdit text
            :returns: The mount point path
            :rtype: str or None
        """
        kf = self.unlock_keyfile.text().strip()
        return kf if kf != '' else None

    def get_mount_point(self):
        """ Getter for QLineEdit text
            :returns: The mount point path
            :rtype: str or None
        """
        mp = self.unlock_mountpoint.text().strip()
        return mp if mp != '' else None

    def show_help_create(self):
        """ Triggered by clicking the help button (create tab) """
        header_text = _('<b>Create a new encrypted container</b>\n')
        basic_help = _('Enter the path of the <b>new container file</b> in the textbox '
                       'or click the button next to the box for a graphical create file dialog.'
                       '\n\n'
                       'The <b>device name</b> will be used to identify the unlocked container. '
                       'It can be any name up to 16 unicode characters, as long as it is unique.'
                       '\n\n'
                       'The <b>size</b> of the container can be provided in GB or MB. The container '
                       'will get initialized with random data, this can take quite a while - '
                       '1 hour for a 10GB container on an external drive is nothing unusual.'
                       '\n\n'
                       'To speed up container creation <b>Quickformat</b> can be enabled to use `fallocate` '
                       'instead of initializing the container with random data - this means previous data '
                       'will not be overwritten and some conclusions about encrypted data inside closed '
                       'containers can be drawn.\n')
        advanced_topics = [
            {'head': _('key file'),
             'text': _('A key file can be used to allow access to an encrypted container instead of a password. '
                       'Using a key file resembles unlocking a door with a key in the real world - anyone with '
                       'access to the key file can open your encrypted container. Make sure to store it at a '
                       'protected location. Its okay to store it on your computer if you are using an already '
                       'encrypted harddrive or a digital keystore. Having the key file on a '
                       '<a href="https://www.google.com/search?q=keychain+usb+drive">small USB drive</a> '
                       'attached to your real chain of keys would be an option as well.\n'
                       'Since you dont have to enter a password, using a key file can be a convenient way to '
                       'access your encrypted container. Just make sure you dont lose the key (file) ;)')
                + _('\n\n'
                    'Although basically any file could be used as a key file, a file with predictable content '
                    'leads to similar problems as using weak passwords. Audio files or pictures are a good choice. '
                    'If unsure use the `create key file` button to generate a small key file '
                    'filled with random data.')},
            {'head': _('encryption format'),
             'text': _('The standard disk encryption format on Linux is called LUKS. '
                       'With <a href="https://github.com/t-d-k/LibreCrypt">LibreCrypt</a> '
                       'you can use LUKS containers on Windows as well. '
                       'The TrueCrypt format is quite popular on Windows/Mac, and can be created '
                       'on Linux if `tcplay` is installed. Please note, that "hidden" TrueCrypt '
                       'partitions are not supported by luckyLUKS!')},
            {'head': _('filesystem'),
             'text': _('Choose the ntfs filesystem to be able to access your data from Linux, '
                       'Windows and Mac OSX. Since access permissions cannot be mapped from '
                       'ntfs to Linux, access to ntfs devices is usually not restricted '
                       '-> take care when using unlocked ntfs devices in a multiuser environment!')}
        ]
        hd = HelpDialog(self, header_text, basic_help, advanced_topics)
        hd.exec_()

    def show_help_unlock(self):
        """ Triggered by clicking the help button (unlock tab) """
        header_text = _('<b>Unlock an encrypted container</b>\n')
        basic_help = _('Select the encrypted <b>container file</b> by clicking the button next to '
                       'the textbox. Both LUKS and Truecrypt containers are supported!'
                       '\n\n'
                       'The <b>device name</b> will be used to identify the unlocked container. '
                       'It can be any name up to 16 unicode characters, as long as it is unique '
                       '-> you cannot give two unlocked containers the same name')
        advanced_topics = [
            {'head': _('key file'),
             'text': _('A key file can be used to allow access to an encrypted container instead of a password. '
                       'Using a key file resembles unlocking a door with a key in the real world - anyone with '
                       'access to the key file can open your encrypted container. Make sure to store it at a '
                       'protected location. Its okay to store it on your computer if you are using an already '
                       'encrypted harddrive or a digital keystore. Having the key file on a '
                       '<a href="https://www.google.com/search?q=keychain+usb+drive">small USB drive</a> '
                       'attached to your real chain of keys would be an option as well.\n'
                       'Since you dont have to enter a password, using a key file can be a convenient way to '
                       'access your encrypted container. Just make sure you dont lose the key (file) ;)')},
            {'head': _('mount point'),
             'text': _('The mount point is the folder on your computer, where you can '
                       'access the files inside the container after unlocking. '
                       'If automatic mounting is configured on your system (eg with udisks), '
                       'explicitly setting a mountpoint is not neccessary (but still possible).')}
        ]
        hd = HelpDialog(self, header_text, basic_help, advanced_topics)
        hd.exec_()



"""
Helper to establish a root-worker process and ipc handler.

luckyLUKS Copyright (c) 2014,2015,2022 Jasper van Hoorn (muzius@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <http://www.gnu.org/licenses/>
"""

import os
import select
import fcntl
import subprocess
import sys
import json
import traceback

from PyQt5.QtCore import QThread, QEvent
from PyQt5.QtWidgets import QApplication

from luckyLUKS.unlockUI import PasswordDialog, SudoDialog, UserInputError
from luckyLUKS.utilsUI import show_alert, show_info


class SudoException(Exception):
    """ Errors while establishing a worker process with elevated privileges using sudo"""


class WorkerMonitor(QThread):

    """ Establishes an asynchronous communication channel with the worker process:
        Since the worker executes only one task at a time queueing/sophisticated ipc are not needed.
        After execute is called with command and callbacks, further commands will get blocked
        until an answer from the worker arrives. The answer will get injected into the UI-loop
        -> the UI stays responsive, and thus has to disable buttons etc to prevent the user from
           sending additional commands
    """

    def __init__(self, parent):
        """ :param parent: The parent widget to be passed to modal dialogs
            :type parent: :class:`PyQt5.QtGui.QWidget`
            :raises: SudoException
        """
        super().__init__()
        self.parent = parent
        self.success_callback, self.error_callback = None, None
        self.modify_sudoers = False
        self.worker = None
        self._spawn_worker()

        if self.modify_sudoers:  # adding user/program to /etc/sudoers.d/ requested
            self.execute({'type': 'request', 'msg': 'authorize'}, None, None)
            response = json.loads(self.worker.stdout.readline().strip())  # blocks
            if response['type'] == 'error':
                show_alert(self.parent, response['msg'])
            else:
                message = _('Permanent `sudo` authorization for\n'
                            '{program}\n'
                            'has been successfully added for user `{username}` to \n'
                            '/etc/sudoers.d/lucky-luks\n').format(
                    program=os.path.abspath(sys.argv[0]),
                    username=os.getenv("USER"))
                show_info(self.parent, message, _('Success'))

    def _spawn_worker(self):
        """ Init worker subprocess with sudo && setup ipc handler
            :raises: SudoException
        """
        # using poll to wait for feedback from sudo
        self.pipe_events = select.poll()
        self._connect_to_sudo()
        dlg_message = _('luckyLUKS needs administrative privileges.\nPlease enter your password:')
        incorrent_pw_entered = False

        try:
            while True:
                __, event = self.pipe_events.poll()[0]  # blocking
                # sudo process wrote to pipe -> read message
                msg = self.worker.stdout.read()

                if event & select.POLLIN:
                    if 'ESTABLISHED' in msg:
                        # Helper process initialized, from here on all com-messages on the pipe
                        # will be terminated with newline -> switch back to blocking IO
                        fl = fcntl.fcntl(self.worker.stdout.fileno(), fcntl.F_GETFL)
                        fcntl.fcntl(self.worker.stdout.fileno(), fcntl.F_SETFL, fl & (~os.O_NONBLOCK))
                        break

                    if 'SUDO_PASSWD_PROMPT' in msg:
                        if incorrent_pw_entered:
                            dlg_msg = _('<b>Sorry, incorrect password.</b>\n') + dlg_message
                        else:
                            dlg_msg = dlg_message
                        self.worker.stdin.write(
                            SudoDialog(parent=self.parent,
                                       message=dlg_msg,
                                       toggle_function=lambda val: setattr(self, 'modify_sudoers', val)
                                       ).get_password() + '\n')
                        self.worker.stdin.flush()
                        incorrent_pw_entered = True

                    elif 'incorrect password attempts' in msg:
                        # max password attempts reached -> restart sudo process and continue
                        self._connect_to_sudo()

                    elif 'not allowed to execute' in msg or 'not in the sudoers file' in msg:
                        dlg_su_message = _('You are not allowed to execute this script with `sudo`.\n'
                                           'If you want to modify your `sudo` configuration,\n'
                                           'please enter the <b>root/administrator</b> password.\n')
                        incorrent_pw_entered = False
                        while True:
                            master, slave = os.openpty()  # su has to be run from a terminal
                            p = subprocess.Popen(
                                "su -c '" + sys.argv[0] + " --ishelperprocess --sudouser " + str(os.getuid()) + "'",
                                shell=True, stdin=slave, stderr=subprocess.PIPE, stdout=subprocess.PIPE,
                                universal_newlines=True, close_fds=True
                            )
                            if incorrent_pw_entered:
                                dlg_msg = _('<b>Sorry, incorrect password.</b>\n') + dlg_su_message
                            else:
                                dlg_msg = dlg_su_message
                            os.write(master, (PasswordDialog(parent=self.parent,
                                                             message=dlg_msg
                                                             ).get_password() + '\n').encode('UTF-8'))
                            p.wait()

                            if p.returncode == 0:
                                show_info(
                                    self.parent,
                                    _('`sudo` configuration successfully modified, now\n'
                                      'you can use luckyLUKS with your user password.\n\n'
                                      'If you want to grant permanent administrative rights\n'
                                      'just tick the checkbox in the following dialog.\n'),
                                    _('Success')
                                )
                                incorrent_pw_entered = False
                                self._connect_to_sudo()
                                break
                            if p.returncode == 1:
                                incorrent_pw_entered = True
                            else:
                                # worker prints exceptions to stdout
                                # to keep them separated from 'su: Authentication failure'
                                raise SudoException(p.stdout.read())

                elif event & select.POLLERR or event & select.POLLHUP:
                    raise SudoException(msg)

        except SudoException:  # don't touch
            raise
        except UserInputError as e:  # user cancelled dlg -> quit without msg
            raise SudoException() from e
        except Exception as e:  # catch ANY other exception to show via gui
            raise SudoException(
                _('Communication with sudo process failed\n{error}')
                .format(error=''.join(traceback.format_exception(*sys.exc_info())))
            ) from e
        finally:
            try:
                self.pipe_events.unregister(self.worker.stdout.fileno())
            except KeyError:
                pass  # fd might already gone (IOError etc)
            del self.pipe_events

    def _connect_to_sudo(self):
        """ Calls worker process with sudo and initializes pipes for communication """
        if self.worker is not None:
            # disconnect event listener and wait for process termination
            self.pipe_events.unregister(self.worker.stdout.fileno())
            self.worker.wait()
        # since output from sudo gets parsed, it needs to be run without localization
        # saving original language settings / LC-environment to pass to the worker process
        original_language = os.getenv("LANGUAGE", "")
        env_lang_cleared = {prop: os.environ[prop] for prop in os.environ if prop[0:3] == 'LC_' or prop == 'LANG'}
        env_lang_cleared['LANGUAGE'] = 'C'
        cmd = ['sudo', '-S', '-p', 'SUDO_PASSWD_PROMPT',
               'LANGUAGE=' + original_language,
               os.path.abspath(sys.argv[0]),
               '--ishelperprocess']
        self.worker = subprocess.Popen(cmd,
                                       stdin=subprocess.PIPE, stderr=subprocess.STDOUT,
                                       stdout=subprocess.PIPE, universal_newlines=True, env=env_lang_cleared)
        # switch pipe to non-blocking IO
        fd = self.worker.stdout.fileno()
        fl = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)
        # connect event listener
        self.pipe_events.register(self.worker.stdout.fileno(), select.POLLIN)

    def run(self):
        """ Listens on workers stdout and executes callbacks when answers arrive """
        while True:
            try:
                buf = self.worker.stdout.readline()  # blocks
                if buf:  # check if worker output pipe closed
                    response = json.loads(buf.strip())
                else:
                    return
                assert('type' in response and 'msg' in response)
                # there should be somebody waiting for an answer!
                assert(self.success_callback is not None and self.error_callback is not None)
                # valid response received
                if response['type'] == 'error':
                    QApplication.postEvent(self.parent, WorkerEvent(self.error_callback, response['msg']))
                else:
                    QApplication.postEvent(self.parent, WorkerEvent(self.success_callback, response['msg']))
                # reset callbacks
                self.success_callback, self.error_callback = None, None

            except ValueError:
                # worker didn't return json -> probably crashed, show everything printed to stdout
                os.set_blocking(self.worker.stdout.fileno(), False)
                buf += str(self.worker.stdout.readlines())
                QApplication.postEvent(
                    self.parent,
                    WorkerEvent(callback=lambda msg: show_alert(self.parent, msg, critical=True),
                                response=_('Error in communication:\n{error}').format(error=_(buf)))
                )
                return

            except (IOError, AssertionError) as communication_error:
                QApplication.postEvent(
                    self.parent,
                    WorkerEvent(callback=lambda msg: show_alert(self.parent, msg, critical=True),
                                response=_('Error in communication:\n{error}').format(error=str(communication_error)))
                )
                return

    def execute(self, command, success_callback, error_callback):
        """ Writes command to workers stdin and sets callbacks for listener thread
            :param command: The function to be done by the worker is in command[`msg`]
                            the arguments are passed as named properties command[`device_name`] etc.
            :type command: dict
            :param success_callback: The function to be called if the worker finished successfully
            :type success_callback: function
            :param error_callback: The function to be called if the worker returns an error
            :type error_callback: function
        """
        try:
            # valid command obj?
            assert('type' in command and 'msg' in command)
            # channel clear? (no qeue neccessary for the backend process)
            assert(self.success_callback is None and self.error_callback is None)
            self.success_callback = success_callback
            self.error_callback = error_callback
            self.worker.stdin.write(json.dumps(command) + '\n')
            self.worker.stdin.flush()
        except (IOError, AssertionError) as communication_error:
            QApplication.postEvent(
                self.parent,
                WorkerEvent(callback=lambda msg: show_alert(self.parent, msg, critical=True),
                            response=_('Error in communication:\n{error}').format(error=str(communication_error)))
            )


class WorkerEvent(QEvent):

    """ thread-safe callback execution by raising
        these custom events in the main ui loop
    """
    EVENT_TYPE = QEvent.Type(QEvent.registerEventType())

    def __init__(self, callback, response):
        """ A WorkerEvent encapsulates a function to be called in the main ui loop and its argument
            :param callback: The function to be called when the event gets processed
            :type callback: function
            :param response: Response message from the worker, passed as argument to the callback function
            :type response: str
        """
        QEvent.__init__(self, WorkerEvent.EVENT_TYPE)
        self.callback = callback
        self.response = response


class KeyfileCreator(QThread):

    """ Create a 1KByte key file with random data
        Worker thread to avoid blocking the ui loop
    """

    def __init__(self, parent, path):
        """ :param parent: The parent widget to be passed to modal dialogs
            :type parent: :class:`PyQt5.QtGui.QWidget`
            :param path: The designated key file path
            :type path: str
        """
        super().__init__()
        self.parent = parent
        self.path = path
        self.process = None

    def run(self):
        """ Spawns child process and passes a WorkerEvent to the main event loop when finished """
        try:
            output_file = str(self.path)
        except UnicodeEncodeError:
            output_file = self.path.encode('utf-8')  # assume uft8 encoding for shell - see worker
        # oflag=excl -> fail if the output file already exists
        cmd = ['dd', 'if=/dev/random', 'of=' + output_file, 'bs=1', 'count=1024', 'conv=excl']
        with open(os.devnull) as DEVNULL:
            self.process = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=DEVNULL,
                                            universal_newlines=True, close_fds=True)
            __, errors = self.process.communicate()
        if self.process.returncode != 0:
            QApplication.postEvent(
                self.parent.parent(),
                WorkerEvent(callback=lambda msg: self.parent.display_create_failed(msg, stop_timer=True),
                            response=_('Error while creating key file:\n{error}').format(error=errors))
            )
        else:
            QApplication.postEvent(
                self.parent.parent(),
                WorkerEvent(callback=lambda msg: self.parent.on_keyfile_created(msg),
                            response=self.path)
            )

    def terminate(self):
        """ kill dd process """
        self.process.kill()


def is_installed(executable):
    """ Checks if executable is present
        Because the executables will be run by the privileged worker process,
        the usual root path gets added to the users environment path.
        Note: an executable at a custom user path will only be used by the worker process,
        if it is also present in the root path -> therefore this check might not be 100% accurate,
        but almost always sufficient. Checking the real root path would require calling
        the worker process, this way in rare cases the worker might throw an error on startup
        :param executable: executable to search for
        :type executable: str
        :returns: True if executable found
        :rtype: bool
    """
    return any([os.path.exists(os.path.join(p, executable))
                for p in os.environ["PATH"].split(os.pathsep) + ['/sbin', '/usr/sbin']
                ])



"""
Because access to /dev/mapper is restricted to root, administrator privileges are needed
to handle encrypted containers as block devices. The GUI will be run from a normal
user account while all calls to cryptsetup and mount will be executed from a separate
worker process with administrator privileges. Everything that runs with elevated privs
is included in this module.

Copyright (c) 2014,2015,2022 Jasper van Hoorn (muzius@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
"""

import subprocess
import os
import json
import sys
import traceback
import pwd
import stat
import warnings
import threading
import signal
import random
from uuid import uuid4
from time import sleep
import queue


class WorkerException(Exception):
    """ Used to catch error messages from shell calls to give feedback via gtk-gui """


class UserAbort(Exception):
    """ Worker gets notified about user canceling the command -> no response needed """


def com_thread(cmdqueue):
    """ Monitors the incoming pipe and puts commands in a queue. Linux signals cannot be sent
        from the parent to a privileged childprocess, so to send sigterm this thread detects if the parent
        closes the pipe instead, to terminate the main thread and all its child processes.
        :param cmdqueue: Queue to pass incoming commands to main thread
        :type cmdqueue: queue
    """
    while True:
        try:
            buf = sys.stdin.readline().strip()  # blocks
            if not buf:  # check if input pipe closed
                break
            cmdqueue.put(json.loads(buf))
        except IOError:
            break
    # send INT to all child processes (eg currently creating new container with dd etc..)
    os.killpg(0, signal.SIGINT)


def run():
    """ Initialize helper and setup ipc. Reads json encoded, newline terminated commands from stdin,
        performs the requested command and returns the json encoded, newline terminated answer on stdout.
        All commands are executed sequentially, but stdin.readline() gets read in another thread that monitors
        the incoming pipe and passes to a queue or quits the helper process if the parent closes the pipe.
        Watching for closed pipe is needed because normal signals cannot be sent to a privileged childprocess """

    if os.getuid() != 0:
        sys.stdout.write(_('Please call with sudo.'))
        sys.exit(1)
    elif os.getenv("SUDO_USER") is None or os.getenv("SUDO_UID") is None or os.getenv("SUDO_GID") is None:
        sys.stdout.write(_('Missing information of the calling user in sudo environment.\n'
                           'Please make sure sudo is configured correctly.'))
        sys.exit(1)
    else:
        # send ack to establish simple json encoded request/response protocol using \n as terminator
        sys.stdout.write('ESTABLISHED')
        sys.stdout.flush()

    # start thread to monitor the incomming pipe, communicate via queue
    cmdqueue = queue.Queue()
    worker = WorkerHelper(cmdqueue)
    t = threading.Thread(target=com_thread, args=(cmdqueue,))
    t.start()
    # create process group to be able to quit all child processes of the worker
    os.setpgrp()

    with warnings.catch_warnings():
        warnings.filterwarnings('error')  # catch warnings to keep them from messing up the pipe
        while True:
            response = {'type': 'response', 'msg': 'success'}  # return success unless exception
            try:
                # arbitrary timeout needed to be able to get instant KeyboardInterrupt
                # pythons queue.get() without timeout prefers not to be interrupted by KeyboardInterrupt :)
                cmd = cmdqueue.get(timeout=32767)
                if cmd['msg'] == 'status':
                    is_unlocked = worker.check_status(cmd['device_name'], cmd['container_path'],
                                                      cmd['key_file'], cmd['mount_point'])
                    if not is_unlocked and cmd['key_file'] is not None:  # if keyfile used try to unlock on startup
                        worker.unlock_container(cmd['device_name'], cmd['container_path'],
                                                cmd['key_file'], cmd['mount_point'])
                        response['msg'] = 'unlocked'
                    else:
                        response['msg'] = 'unlocked' if is_unlocked else 'closed'
                elif cmd['msg'] == 'unlock':
                    worker.unlock_container(cmd['device_name'], cmd['container_path'],
                                            cmd['key_file'], cmd['mount_point'])
                elif cmd['msg'] == 'close':
                    worker.close_container(cmd['device_name'], cmd['container_path'])
                elif cmd['msg'] == 'create':
                    worker.create_container(cmd['device_name'], cmd['container_path'],
                                            cmd['container_size'], cmd['filesystem_type'],
                                            cmd['encryption_format'], cmd['key_file'],
                                            cmd['quickformat'])
                elif cmd['msg'] == 'authorize':
                    worker.modify_sudoers(os.getenv("SUDO_UID"), nopassword=True)
                else:
                    raise WorkerException(_('Helper process received unknown command'))
            except queue.Empty:
                continue  # timeout reached on queue.get() -> keep polling
            except UserAbort:
                continue  # no response needed
            except WorkerException as we:
                response = {'type': 'error', 'msg': str(we)}
            except KeyError as ke:
                # thrown if required parameters missing
                response = {'type': 'error', 'msg': _('Error in communication:\n{error}').format(error=str(ke))}
            except KeyboardInterrupt:
                # gets raised by killpg -> quit
                sys.exit(0)
            except Exception:  # catch ANY exception (including warnings) to show via gui
                response = {'type': 'error', 'msg': ''.join(traceback.format_exception(*sys.exc_info()))}
            sys.stdout.write(json.dumps(response) + '\n')
            sys.stdout.flush()


class WorkerHelper():

    """ accepts 5 commands:
        -> check_status() validates the input and returns the current state (unlocked/closed) of the container
        -> unlock_container() asks for the passphrase and tries to unlock and mount a container
        -> close_container() closes and unmounts a container
        -> create_container() initializes a new encrypted LUKS container and sets up the filesystem
        -> modify_sudoers() adds sudo access to the program without password for the current user (/etc/sudoers.d/)
    """

    def __init__(self, cmdqueue=None):
        """ Check tcplay installation """
        self.cmdqueue = cmdqueue
        self.is_tc_installed = any([os.path.exists(os.path.join(p, 'tcplay'))
                                    for p in os.environ["PATH"].split(os.pathsep)])

    def communicate(self, request):
        """ Helper to get an synchronous response from UI (obtain Passphrase or signal create progress)
            :param request: message to send to the UI
            :type request: str
            :returns: The JSON encoded response from the UI
            :rtype: str
            :raises: UserAbort
        """
        sys.stdout.write(json.dumps({'type': 'request', 'msg': request}) + '\n')
        sys.stdout.flush()
        response = self.cmdqueue.get()  # wait for response
        try:
            assert('type' in response and 'msg' in response)
        except AssertionError as ae:
            raise UserAbort() from ae
        # quit if abort or unexpected msg from ui
        if response['type'] != 'response':
            raise UserAbort()

        return response['msg']

    def check_status(self, device_name, container_path, key_file=None, mount_point=None):
        """
            Validates the input and returns the current state (unlocked/closed) of the container.
            The checks are sufficient to keep users from shooting themselves in the foot and
            provide the most possible protection against TOCTOU attacks: Since most commands are
            executed using the device mapper name, there is not much attack surface for TOCTOU anyway,
            since root access is needed to access device mapper. It would be possible to unmount
            other users containers with the right timing though - slightly annoying but no serious threat
            :param device_name: The device mapper name
            :type device_name: str
            :param container_path: The path of the container file
            :type container_path: str
            :param key_file: The path to an optional keyfile to be used for the container
            :type key_file: str or None
            :param mount_point: The path of an optional mount point
            :type mount_point: str or None
            :returns: True if LUKS device is active/unlocked
            :rtype: bool
            :raises: WorkerException
        """
        uid = int(os.getenv("SUDO_UID"))

        # device_name and container_path valid?
        if device_name == '':
            raise WorkerException(_('Device Name is empty'))
        # check access rights to container file
        if not os.path.exists(container_path) or os.stat(container_path).st_uid != uid:
            sleep(random.random())  # 0-1s to prevent misuse of exists()
            raise WorkerException(
                _('Container file not accessible\nor path does not exist:\n\n{file_path}')
                .format(file_path=container_path)
            )

        is_unlocked = self.is_LUKS_active(device_name)

        if is_unlocked:
            # make sure container file currently in use for device name is the same as the supplied container path
            if container_path != self.get_container(device_name):
                raise WorkerException(_('Could not use container:\n{file_path}\n'
                                        '<b>{device_name}</b> is already unlocked\n'
                                        'using a different container\n'
                                        'Please change the name to unlock this container')
                                      .format(file_path=container_path, device_name=device_name))

        # container is not unlocked
        else:
            # validate device name
            if len(bytes(device_name, 'utf-8')) > 16:
                # the goal here is not to confuse the user:
                # thus Name==Label of the partition inside the encrypted container
                # -> because thats what usually gets presented to the user (filemanager)
                # see checks in create_container below for length restictions on partition labels
                # dev/mapper would only be able to handle slightly longer names for unicode anyways
                raise WorkerException(
                    _('Device Name too long:\n'
                      'Only up to 16 characters possible, even less for unicode\n'
                      '(roughly 8 non-ascii characters possible)')
                )

            if device_name[0:1] == '-' or '/' in device_name:
                # cryptsetup thinks -isanoption and "/" is not supported by devmapper
                raise WorkerException(_('Illegal Device Name!\nNames starting with `-` or using `/` are not possible'))

            # prevent container from being unlocked multiple times with different names
            container_found = subprocess.check_output(
                ['losetup', '-j', container_path],
                stderr=subprocess.STDOUT, universal_newlines=True
            )
            if container_found != '':
                # container is already in use -> try to find out the device name
                existing_device_name = ''
                encrypted_devices = subprocess.check_output(
                    ['dmsetup', 'status', '--target', 'crypt'],
                    stderr=subprocess.STDOUT, universal_newlines=True
                ).strip()
                for encrypted_device in encrypted_devices.split('\n'):
                    encrypted_device = encrypted_device[:encrypted_device.find(':')].strip()
                    if container_path == self.get_container(encrypted_device):
                        existing_device_name = encrypted_device
                        break
                raise WorkerException(
                    _('Cannot use the container\n'
                      '{file_path}\n'
                      'The container is already in use ({existing_device}).')
                    .format(file_path=container_path, existing_device=existing_device_name)
                )

            # validate key_file if given
            if (key_file is not None and
                any([(not os.path.exists(key_file)),
                     (os.stat(key_file).st_uid != int(os.getenv("SUDO_UID")))])):
                sleep(random.random())  # 0-1s to prevent misuse of exists()
                raise WorkerException(
                    _('Key file not accessible\nor path does not exist:\n\n{file_path}')
                    .format(file_path=key_file))

            # validate mount_point if given
            if mount_point is not None:

                if not os.path.exists(mount_point) or os.stat(mount_point).st_uid != uid:
                    sleep(random.random())  # 0-1s to prevent misuse of exists()
                    raise WorkerException(
                        _('Mount point not accessible\nor path does not exist:\n\n{mount_dir}')
                        .format(mount_dir=mount_point)
                    )
                if os.path.ismount(mount_point):
                    raise WorkerException(
                        _('Already mounted at mount point:\n\n{mount_dir}')
                        .format(mount_dir=mount_point)
                    )
                if os.listdir(mount_point):
                    raise WorkerException(
                        _('Designated mount directory\n{mount_dir}\nis not empty')
                        .format(mount_dir=mount_point)
                    )

        return is_unlocked

    def unlock_container(self, device_name, container_path, key_file=None, mount_point=None, pw_callback=None):
        """ Unlocks LUKS or Truecrypt containers.
            Validates input and keeps asking
            for the passphrase until successfull unlock,
            followed by an optional mount.
            :param device_name: The device mapper name
            :type device_name: str
            :param container_path: The path of the container file
            :type container_path: str
            :param key_file: The path to an optional keyfile to be used for the container
            :type key_file: str or None
            :param mount_point: The path of an optional mount point
            :type mount_point: str or None
            :param pw_callback: A callback function that returns the password for unlocking
            :type pw_callback: function()
            :raises: WorkerException
        """
        is_unlocked = self.check_status(device_name, container_path, key_file, mount_point)
        if not is_unlocked:  # just return if unlocked -> does not mount an already unlocked container
            if pw_callback is None:
                pw_callback = lambda: self.communicate('getPassword')
            # workaround udisks-daemon crash (udisksd from udisks2 is okay): although cryptsetup is able to handle
            # loopback device creation/teardown itself, using this crashes udisks-daemon
            # -> manual loopback device handling here
            # TODO: could be removed, udisks is replaced with udisks2 since ~2016
            try:
                loop_dev = subprocess.check_output(
                    ['losetup', '-f', '--show', container_path],
                    stderr=subprocess.PIPE, universal_newlines=True
                ).strip()
            except subprocess.CalledProcessError as cpe:
                # most likely no more loopdevices available
                raise WorkerException(cpe.output) from cpe
            crypt_initialized = False

            try:
                # check if LUKS container, try Truecrypt otherwise (tc container cannot be identified by design)
                container_is_luks = (subprocess.call(['cryptsetup', 'isLuks', container_path]) == 0)
                if container_is_luks:
                    open_command = ['cryptsetup', 'open', loop_dev, device_name]
                else:
                    open_command = ['cryptsetup', 'open', '--type', 'tcrypt', loop_dev, device_name]

                with open(os.devnull) as DEVNULL:
                    if key_file is None:
                        while not is_unlocked:
                            p = subprocess.Popen(
                                open_command,
                                stdin=subprocess.PIPE, stderr=subprocess.PIPE, stdout=DEVNULL,
                                universal_newlines=True, close_fds=True
                            )
                            __, errors = p.communicate(pw_callback())
                            if p.returncode == 0:
                                is_unlocked = True
                            elif p.returncode == 2:  # cryptsetup: no permission (bad passphrase)
                                continue
                            else:
                                raise WorkerException(errors)
                    else:
                        open_command += ['--key-file', key_file]
                        p = subprocess.Popen(
                            open_command,
                            stdin=subprocess.PIPE, stderr=subprocess.PIPE, stdout=DEVNULL,
                            universal_newlines=True, close_fds=True
                        )
                        if container_is_luks:
                            __, errors = p.communicate()
                        else:
                            # tcplay with keyfile only means empty password
                            __, errors = p.communicate('\n')
                        if p.returncode != 0:
                            if p.returncode == 2:
                                # error message from cryptsetup is a bit ambiguous
                                raise WorkerException(_('Open container failed.\nPlease check key file'))
                            raise WorkerException(errors)
                crypt_initialized = True
            finally:
                if not crypt_initialized:
                    self.detach_loopback_device(loop_dev)

            if mount_point is not None:  # only mount if optional parameter mountpoint is set
                try:
                    subprocess.check_output(
                        ['mount', '-o', 'nosuid,nodev', self.get_device_mapper_name(device_name), mount_point],
                        stderr=subprocess.STDOUT, universal_newlines=True)
                except subprocess.CalledProcessError as cpe:
                    raise WorkerException(cpe.output) from cpe

    def close_container(self, device_name, container_path):
        """ Validates input and tries to unmount /dev/mapper/<name> and close container
            :param device_name: The device mapper name
            :type device_name: str
            :param container_path: The path of the container file
            :type container_path: str
            :raises: WorkerException
        """
        if self.check_status(device_name, container_path):  # just return if not unlocked
            # for all mounting /dev/mapper/device_name is used
            try:
                # unfortunately, umount returns the same errorcode for device not mounted and device busy
                # clear locale to be able to parse output
                env_lang_cleared = os.environ.copy()
                env_lang_cleared['LANGUAGE'] = 'C'
                subprocess.check_output(
                    ['umount', self.get_device_mapper_name(device_name)],
                    stderr=subprocess.STDOUT, universal_newlines=True, env=env_lang_cleared
                )
            except subprocess.CalledProcessError as cpe:
                if 'not mounted' not in cpe.output:  # ignore if not mounted and proceed with closing the container
                    raise WorkerException(_('Unable to close container, device is busy')) from cpe
            # get reference to loopback device before closing the container
            associated_loop = self.get_loopback_device(device_name)
            try:
                subprocess.check_output(
                    ['cryptsetup', 'close', device_name],
                    stderr=subprocess.STDOUT, universal_newlines=True
                )
            except subprocess.CalledProcessError as cpe:
                raise WorkerException(cpe.output) from cpe
            # remove loopback device
            sleep(0.2)  # give udisks some time to process closing of container ..
            self.detach_loopback_device(associated_loop)

    def create_container(self, device_name, container_path, container_size,
                         filesystem_type, enc_format, key_file=None, quickformat=False):
        """ Creates a new LUKS2 container with requested size and filesystem after validating parameters
            Three step process: asks for passphrase after initializing container with random bits,
            and signals successful LUKS initialization before writing the filesystem
            :param device_name: The device mapper name, used as filesystem label as well
            :type device_name: str
            :param container_path: The path of the container file to be created
            :type container_path: str
            :param container_size: The size the new container in KB
            :type container_size: int
            :param filesystem_type: The type of the filesystem inside the new container
                                    (supported: 'ext4', 'ext2', 'ntfs')
            :type filesystem_type: str
            :param enc_format: The type of the encryption format used for the new container
                               (supported: 'LUKS', 'TrueCrypt')
            :type enc_format: str
            :param key_file: The path to an optional keyfile to be used for the container
            :type key_file: str or None
            :param quickformat: Use fallocate instead of initializing container with random data
            :type quickformat: bool
            :raises: WorkerException
        """
        # STEP0: #########################################################################
        # Sanitize user input and perform some checks before starting the process to avoid
        # failure later on eg. dd writing out random data for 3 hours to initialize the
        # new container, only to find out that there is not enough space on the device,
        # or that the designated device name is already in use on /dev/mapper
        #

        # validate container file
        if os.path.basename(container_path).strip() == '':
            raise WorkerException(_('Container Filename not supplied'))

        # validate device name
        if device_name == '':
            raise WorkerException(_('Device Name is empty'))

        if device_name[0:1] == '-' or '/' in device_name:
            # cryptsetup thinks -isanoption and "/" is not supported by devmapper
            raise WorkerException(_('Illegal Device Name!\nNames starting with `-` or using `/` are not possible'))

        if len(bytes(device_name, 'utf-8')) > 16:  # ext-labels are the most restricted (max 16Bytes)
            # dev-mapper and ntfs-label would support slightly longer strings..
            raise WorkerException(_('Device Name too long:\n'
                                    'Only up to 16 characters possible, even less for unicode \n'
                                    '(roughly 8 non-ascii characters possible)'))

        if os.path.exists(self.get_device_mapper_name(device_name)):
            raise WorkerException(_('Device Name already in use:\n\n{device_name}').format(device_name=device_name))

        # validate container size
        if container_size < 5242880:
            raise WorkerException(_('Container size too small\n'
                                    'to create encrypted filesystem\n'
                                    'Please choose at least 5MB'))

        container_dir = os.path.dirname(container_path)

        if not os.path.dirname(container_dir):
            container_dir = os.path.expanduser('~' + os.getenv("SUDO_USER"))
            container_path = os.path.join(container_dir, os.path.basename(container_path))

        free_space = os.statvfs(container_dir)
        free_space = free_space.f_bavail * free_space.f_bsize
        if container_size > free_space:
            raise WorkerException(
                _('Not enough free disc space for container:\n\n'
                  '{space_needed} MB needed\n{space_available} MB available')
                .format(space_needed=str(int(container_size / 1024 / 1024)),
                        space_available=str(int(free_space / 1024 / 1024)))
            )

        # validate key_file if given
        if key_file is not None:
            # check access rights to keyfile
            if not os.path.exists(key_file) or os.stat(key_file).st_uid != int(os.getenv("SUDO_UID")):
                sleep(random.random())  # 0-1s to prevent misuse of exists()
                raise WorkerException(
                    _('Key file not accessible\nor path does not exist:\n\n{file_path}').format(file_path=key_file)
                )

        # validate encryption_format and filesystem
        # TODO: exFAT?
        if filesystem_type not in ['ext4', 'ext2', 'ntfs']:
            raise WorkerException(
                _('Unknown filesystem type: {filesystem_type}').format(filesystem_type=str(filesystem_type))
            )
        if enc_format == 'Truecrypt' and not self.is_tc_installed:
            raise WorkerException(_('If you want to use TrueCrypt containers\n'
                                    'make sure `cryptsetup` is at least version 1.6 (`cryptsetup --version`)\n'
                                    'and `tcplay` is installed (eg for Debian/Ubuntu `apt-get install tcplay`)'))

        # STEP1: ##########################################################
        # create container file by filling allocated space with random bits
        #

        # runas user to fail on access restictions
        if quickformat:
            # does not fail if the output file already exists, but check is in setupUI.on_save_file() anyway
            cmd = ['sudo', '-u', os.getenv("SUDO_USER"),
                   'fallocate', '-x', '-l', str(container_size), container_path]
        else:
            count = str(int(container_size / 1024 / 1024)) + 'K'
            # oflag=excl -> fail if the output file already exists
            cmd = ['sudo', '-u', os.getenv("SUDO_USER"),
                   'dd', 'if=/dev/urandom', 'of=' + container_path,
                   'bs=1K', 'count=' + count, 'conv=excl']

        with open(os.devnull) as DEVNULL:
            p = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=DEVNULL, universal_newlines=True, close_fds=True)
            __, errors = p.communicate()
        if p.returncode != 0:
            # 'sudo -u' might add this -> don't display
            raise WorkerException(errors.replace('Sessions still open, not unmounting', '').strip())
            # TODO: this can only work with english locale,
            # but this errormessage doesn't seem to be localized in sudo yet ..
            # get rid of the problem (strip env?) or remove msg in all languages

        # setup loopback device with created container
        try:
            reserved_loopback_device = subprocess.check_output(
                ['losetup', '-f', '--show', container_path],
                stderr=subprocess.PIPE, universal_newlines=True
            ).strip()
        except subprocess.CalledProcessError as cpe:
            raise WorkerException(cpe.output) from cpe

        # STEP2: ######################################################
        # ask user for password and initialize LUKS/TrueCrypt container
        #
        resp = ''
        try:
            if key_file is None:
                resp = self.communicate('getPassword')
            else:
                self.communicate('containerDone')

            if enc_format == 'LUKS':

                cmd = ['cryptsetup', 'luksFormat', '--type', 'luks2', '-q', reserved_loopback_device]
                if key_file is not None:
                    cmd += ['--key-file', key_file]
                with open(os.devnull) as DEVNULL:
                    p = subprocess.Popen(cmd,
                                         stdin=subprocess.PIPE, stderr=subprocess.PIPE, stdout=DEVNULL,
                                         universal_newlines=True, close_fds=True)
                    __, errors = p.communicate(resp)
                if p.returncode != 0:
                    raise WorkerException(errors)

            elif enc_format == 'TrueCrypt':

                with open(os.devnull) as DEVNULL:
                    # secure erase already done with dd, no need to use tcplay for that
                    cmd = ['tcplay', '-c', '-d', reserved_loopback_device, '--insecure-erase']
                    if key_file is not None:
                        cmd += ['--keyfile', key_file]
                    p = subprocess.Popen(cmd,
                                         stdin=subprocess.PIPE, stderr=subprocess.PIPE, stdout=DEVNULL,
                                         universal_newlines=True, close_fds=True)
                    # tcplay needs the password twice & confirm -> using sleep instead of parsing output
                    # ugly, but crypt-init takes ages with truecrypt anyways
                    sleep(1)
                    p.stdin.write(resp + '\n')
                    p.stdin.flush()
                    sleep(1)
                    p.stdin.write(resp + '\n')
                    p.stdin.flush()
                    sleep(1)
                    p.stdin.write('y\n')  # .. until tcplay gets localized :)
                    p.stdin.flush()
                    p.stdin.close()
                    p.stderr.close()
                    p.wait()

                if p.returncode != 0:
                    raise WorkerException('TCPLAY ERROR')

            else:
                raise WorkerException(_('Unknown encryption format: {enc_fmt}').format(enc_fmt=enc_format))
            self.communicate('formatDone')  # signal status
        finally:  # cleanup loopback device
            self.detach_loopback_device(reserved_loopback_device)

        # STEP3: ############################################
        # open encrypted container and format with filesystem
        #
        pw_callback = lambda: resp
        self.unlock_container(device_name=device_name,
                              container_path=container_path,
                              key_file=key_file,
                              pw_callback=pw_callback)
        resp = None  # get rid of pw

        # fs-root of created ext-filesystem should belong to the user
        device_mapper_name = self.get_device_mapper_name(device_name)
        if filesystem_type == 'ext4':
            cmd = ['mkfs.ext4', '-L', device_name, '-O', '^has_journal', '-m', '0', '-q', device_mapper_name]
        elif filesystem_type == 'ext2':
            cmd = ['mkfs.ext2', '-L', device_name, '-m', '0', '-q', device_mapper_name]
        elif filesystem_type == 'ntfs':
            cmd = ['mkfs.ntfs', '-L', device_name, '-Q', '-q', device_mapper_name]
        try:
            subprocess.check_output(cmd, stderr=subprocess.STDOUT, universal_newlines=True)
        except subprocess.CalledProcessError as cpe:
            raise WorkerException(cpe.output) from cpe

        # remove group/other read/execute rights from fs root if possible
        if filesystem_type != 'ntfs':
            tmp_mount = os.path.join('/tmp/', str(uuid4()))
            os.mkdir(tmp_mount)
            try:
                subprocess.check_output(
                    ['mount', '-o', 'nosuid,nodev', device_mapper_name, tmp_mount],
                    stderr=subprocess.STDOUT, universal_newlines=True
                )
            except subprocess.CalledProcessError as cpe:
                raise WorkerException(cpe.output) from cpe
            os.chown(tmp_mount, int(os.getenv("SUDO_UID")), int(os.getenv("SUDO_GID")))
            os.chmod(tmp_mount, 0o700)

        self.close_container(device_name, container_path)

    def modify_sudoers(self, user_id, nopassword=False):
        """ Adds sudo access to the program (without password) for the current user (/etc/sudoers.d/)
            :param user_id: unix user id
            :type user_id: int
            :returns: should access be granted without password
            :rtype: bool
        """
        program_path = os.path.abspath(sys.argv[0])
        try:
            user_name = pwd.getpwuid(int(user_id))[0]
        except KeyError as ke:
            raise WorkerException(_('Cannot change sudo rights, invalid username')) from ke

        if any([(os.stat(program_path).st_uid != 0),
                (os.stat(program_path).st_gid != 0),
                (os.stat(program_path).st_mode & stat.S_IWOTH)]):

            raise WorkerException(_('I`m afraid I can`t do that.\n\n'
                                    'To be able to permit permanent changes to sudo rights,\n'
                                    'please make sure the program is owned by root\n'
                                    'and not writeable by others.\n'
                                    'Execute the following commands in your shell:\n\n'
                                    'chmod 755 {program}\n'
                                    'sudo chown root:root {program}\n\n').format(program=program_path))

        if not os.path.exists('/etc/sudoers.d'):
            os.makedirs('/etc/sudoers.d')

        sudoers_file_path = '/etc/sudoers.d/lucky-luks'
        sudoers_file = open(sudoers_file_path, 'a')
        sudoers_file.write(
            "{username} ALL = (root) {nopasswd}{program}\n".format(
                username=user_name,
                nopasswd='NOPASSWD: ' if nopassword else '',
                program=program_path
            )
        )
        sudoers_file.close()

        os.chmod(sudoers_file_path, 0o440)

    def is_LUKS_active(self, device_name):
        """ Checks if device is active/unlocked
            :param device_name: The device mapper name
            :type device_name: str
            :returns: True if active LUKS device found
            :rtype: bool
        """
        with open(os.devnull) as DEVNULL:
            returncode = subprocess.call(
                ['cryptsetup', 'status', device_name],
                stdout=DEVNULL, stderr=subprocess.STDOUT
            )
        return returncode == 0

    def detach_loopback_device(self, loopback_device):
        """ Detaches given loopback device
            :param loopback_device: The loopback device path (eg /dev/loop2)
            :type loopback_device: str
        """
        with open(os.devnull) as DEVNULL:
            subprocess.call(['losetup', '-d', loopback_device], stdout=DEVNULL, stderr=subprocess.STDOUT)

    def get_loopback_device(self, device_name):
        """ Returns the corresponding loopback device path to a given device mapper name
            :param device_name: The device mapper name
            :type device_name: str
            :returns: The corresponding loopback device path (eg /dev/loop2)
            :rtype: str
        """
        return self.get_crypt_status(device_name, 'device:')

    def get_container(self, device_name):
        """ Returns the corresponding container path to a given device mapper name
            :param device_name: The device mapper name
            :type device_name: str
            :returns: The corresponding container path
            :rtype: str
        """
        return self.get_crypt_status(device_name, 'loop:')

    def get_crypt_status(self, device_name, search_property):
        """ Parses cryptsetup status output for a device mapper name
            and returns either loopback device or container path
            :param device_name: The device mapper name
            :type device_name: str
            :param search_property: The property to return from status output
            :type search_property: 'device:' or 'loop:'
            :returns: loopback device or container path
            :rtype: str
        """
        try:
            stat_output = subprocess.check_output(
                ['cryptsetup', 'status', device_name],
                stderr=subprocess.STDOUT, universal_newlines=True
            )
            # parsing status output: should be safe, output unchanged in cryptsetup since version 1.3 (2011)
            for line in stat_output.split('\n'):
                if search_property in line:
                    return line[line.find('/'):].strip()
            return ''
        except subprocess.CalledProcessError:
            return ''  # device not found

    def get_device_mapper_name(self, device_name):
        """ Mapping for filesystem access to /dev/mapper/
            Escapes most non alphanumeric and all unicode characters in the device name

            -> from the device-mapper sources:
            * Mangle all characters in the input string which are not on a whitelist
            * with '\xCC' format, where CC is the hex value of the character.
            * Actually, DM supports any character in a device name.
            * This whitelist is just for proper integration with udev.

            :param device_name: The device mapper name
            :type device_name: str
            :returns: The mangled device mapper name for filesystem access
            :rtype: str
        """
        dm_name = '/dev/mapper/'
        # use single-char list for python3 (splitting 2 or more byte utf8-encoded chars..)
        for char in [chr(byte) for byte in bytes(device_name, 'utf-8')]:
            if any([(char >= '0' and char <= '9'),
                    (char >= 'A' and char <= 'Z'),
                    (char >= 'a' and char <= 'z'),
                    (char in "#+-.:=@_")]):
                dm_name += char
            else:
                dm_name += '\\' + str(hex(ord(char))[1:])

        return dm_name



"""
This module contains 3 different graphical modal dialogs to ask a user for password/-phrase.
Each is based on a common password dialog and offers a method to use the dialog
in a synchronous way ie to run itself and return a result or perform an action,
or throw an exception if this fails

luckyLUKS Copyright (c) 2014,2015,2022 Jasper van Hoorn (muzius@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <http://www.gnu.org/licenses/>
"""

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QMessageBox, QDialogButtonBox, QStyle, \
    QLabel, QVBoxLayout, QHBoxLayout, QLineEdit, QCheckBox, QLayout, QApplication
from PyQt5.QtGui import QIcon


class UserInputError(Exception):
    """ Raised if user cancels a password dialog """


class PasswordDialog(QDialog):

    """ Basic dialog with a textbox input field for the password/-phrase and OK/Cancel buttons """

    def __init__(self, parent, message, title='Enter Password'):
        """
            :param parent: The parent window/dialog used to enable modal behaviour
            :type parent: :class:`PyQt5.QtGui.QWidget`
            :param message: The message that gets displayed above the textbox
            :type message: str
            :param title: Displayed in the dialogs titlebar
            :type title: str or None
        """
        super().__init__(parent, Qt.WindowCloseButtonHint | Qt.WindowTitleHint)
        self.setWindowTitle(title)
        self.layout = QVBoxLayout()
        self.layout.setSizeConstraint(QLayout.SetFixedSize)
        self.layout.setSpacing(10)

        # create icon and label
        self.header_box = QHBoxLayout()
        self.header_box.setSpacing(10)
        self.header_box.setAlignment(Qt.AlignLeft)
        self.header_text = QLabel(message)
        icon = QLabel()
        icon.setPixmap(
            QIcon.fromTheme(
                'dialog-password',
                QApplication.style().standardIcon(QStyle.SP_DriveHDIcon)
            ).pixmap(32)
        )
        self.header_box.addWidget(icon)
        self.header_box.addWidget(self.header_text)
        self.layout.addLayout(self.header_box)

        # create the text input field
        self.pw_box = QLineEdit()
        self.pw_box.setMinimumSize(0, 25)
        # password will not be shown on screen
        self.pw_box.setEchoMode(QLineEdit.Password)
        self.layout.addWidget(self.pw_box)
        self.layout.addSpacing(10)

        # OK and Cancel buttons
        self.buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, parent=self)
        self.buttons.accepted.connect(self.on_accepted)
        self.buttons.rejected.connect(self.reject)
        self.layout.addWidget(self.buttons)

        self.setLayout(self.layout)
        self.pw_box.setFocus()

    def on_accepted(self):
        """ Event handler for accept signal: block when input field empty """
        if self.pw_box.text() != '':
            self.accept()
        else:
            self.pw_box.setFocus()

    def get_password(self):
        """ Dialog runs itself and returns the password/-phrase entered
            or throws an exception if the user cancelled/closed the dialog
            :returns: The entered password/-phrase
            :rtype: str
            :raises: UserInputError
        """
        try:
            if self.exec_():
                return str(self.pw_box.text())
            raise UserInputError()
        finally:
            self.destroy()


class SudoDialog(PasswordDialog):

    """ Modified PasswordDialog that adds a checkbox asking for permanent sudo access permission"""

    def __init__(self, parent, message, toggle_function):
        """ :param parent: The parent window/dialog used to enable modal behaviour
            :type parent: :class:`PyQt5.QtGui.QWidget`
            :param message: The message that gets displayed above the textbox
            :type message: str
            :param toggle_function: A function that accepts a boolean value, propagates current state of the checkbox
            :type toggle_function: function(boolean)
        """
        super().__init__(parent, message=message, title=_('luckyLUKS'))

        # add checkbox to dialog
        self.toggle_function = toggle_function
        self.store_cb = QCheckBox('')  # QCheckBox supports only string labels ..
        # using QLabel because markup is needed for linebreak
        cb_label = QLabel(_('Always allow luckyLUKS to be run\nwith administrative privileges'))
        # connect clicked on QLabel to fully emulate QCheckbox behaviour
        cb_label.mouseReleaseEvent = self.toggle_cb
        self.store_cb.stateChanged.connect(self.on_cb_toggled)
        self.toggle_function(False)  # allowing permanent access has to be confirmed explicitly

        self.sudo_box = QHBoxLayout()
        self.sudo_box.setSpacing(5)
        self.sudo_box.addWidget(self.store_cb)
        self.sudo_box.addWidget(cb_label)
        self.sudo_box.addStretch()
        self.layout.insertLayout(2, self.sudo_box)

    def toggle_cb(self, event):
        """ Slot for QCheckbox behaviour emulation: toggles checkbox """
        self.store_cb.setChecked(not self.store_cb.isChecked())

    def on_cb_toggled(self, state):
        """ Event handler for checkbox toggle: propagate new value to parent """
        self.toggle_function(self.store_cb.isChecked())


class FormatContainerDialog(PasswordDialog):

    """ Modified PasswordDialog that shows the input on screen and requests confirmation for close/cancel """

    def __init__(self, parent):
        """ :param parent: The parent window/dialog used to enable modal behaviour
            :type parent: :class:`PyQt5.QtGui.QWidget`
        """
        super().__init__(parent,
                         message=_('Please choose a passphrase\nto encrypt the new container:\n'),
                         title=_('Enter new Passphrase'))
        # display passphrase checkbox and set default to show input
        self.show_cb = QCheckBox(_('Display passphrase'))
        self.show_cb.stateChanged.connect(self.on_cb_toggled)
        self.show_box = QHBoxLayout()
        self.show_box.addWidget(self.show_cb)
        self.layout.insertLayout(2, self.show_box)
        self.show_cb.setChecked(True)

    def on_cb_toggled(self):
        """ Event handler for checkbox toggle: show/hide passphrase input on screen """
        if self.show_cb.isChecked():
            # show input on screen
            self.pw_box.setEchoMode(QLineEdit.Normal)
        else:
            # hide input on screen
            self.pw_box.setEchoMode(QLineEdit.Password)

    def closeEvent(self, event):
        """ Event handler confirm close """
        if not self.confirm_close_cancel():
            event.ignore()

    def reject(self):
        """ Event handler confirm cancel """
        if self.confirm_close_cancel():
            super().reject()
        else:
            self.pw_box.setFocus()

    def confirm_close_cancel(self):
        """ Display dialog for close/cancel confirmation
            :returns: The users decision
            :rtype: bool
        """
        message = _('Currently creating new container!\nDo you really want to quit?')
        mb = QMessageBox(QMessageBox.Question, '', message, QMessageBox.Ok | QMessageBox.Cancel, self)
        mb.button(QMessageBox.Ok).setText(_('Quit'))
        return mb.exec_() == QMessageBox.Ok


class UnlockContainerDialog(PasswordDialog):

    """ Modified PasswordDialog that communicates with the worker process to unlock an encrypted container """

    def __init__(self, parent, worker, luks_device_name, encrypted_container, key_file=None, mount_point=None):
        """ :param parent: The parent window/dialog used to enable modal behaviour
            :type parent: :class:`PyQt5.QtGui.QWidget`
            :param worker: Communication handler with the worker process
            :type worker: :class:`helper.WorkerMonitor`
            :param luks_device_name: The device mapper name
            :type luks_device_name: str
            :param encrypted_container: The path of the container file
            :type encrypted_container: str
            :param key_file: The path of an optional key file
            :type key_file: str or None
            :param mount_point: The path of an optional mount point
            :type mount_point: str or None
        """
        super().__init__(parent, _('Initializing ..'), luks_device_name)

        self.worker = worker
        self.error_message = ''

        if key_file is not None:
            self.header_text.setText(
                _('<b>Using keyfile</b>\n{keyfile}\nto open container.\n\nPlease wait ..')
                .format(keyfile=key_file)
            )
            self.pw_box.hide()
            self.buttons.hide()
            self.header_text.setContentsMargins(10, 10, 10, 10)
        else:
            self.buttons.button(QDialogButtonBox.Ok).setText(_('Unlock'))
            # disable input until worker initialized
            self.pw_box.setEnabled(False)
            self.buttons.setEnabled(False)

        self.waiting_for_response = True
        # call worker
        self.worker.execute(command={'type': 'request',
                                     'msg': 'unlock',
                                     'device_name': luks_device_name,
                                     'container_path': encrypted_container,
                                     'mount_point': mount_point,
                                     'key_file': key_file
                                     },
                            success_callback=self.on_worker_reply,
                            error_callback=self.on_error)

    def on_accepted(self):
        """ Event handler send password/-phrase if worker ready """
        # dont send empty password
        if self.pw_box.text() == '':
            self.pw_box.setFocus()
        elif not self.waiting_for_response:
            # worker is ready, send request
            self.waiting_for_response = True
            self.pw_box.setEnabled(False)
            self.buttons.setEnabled(False)
            self.header_text.setText(_('Checking passphrase ..'))
            self.worker.execute(command={'type': 'response',
                                         'msg': str(self.pw_box.text())
                                         },
                                success_callback=self.on_worker_reply,
                                error_callback=self.on_error)

    def reject(self):
        """ Event handler cancel:
            Block while waiting for response or notify worker with abort message
        """
        if not self.waiting_for_response:
            self.worker.execute({'type': 'abort', 'msg': ''}, None, None)
            super().reject()

    def on_error(self, error_message):
        """ Error-Callback: set errormessage and trigger Cancel """
        self.error_message = error_message
        super().reject()

    def on_worker_reply(self, message):
        """ Success-Callback: trigger OK when unlocked, reset dialog if not """
        if message == 'success':
            self.accept()
        else:
            if self.pw_box.text() == '':  # init
                self.header_text.setText(_('Please enter\ncontainer passphrase:'))
            else:  # at least one previous pw attempt
                self.header_text.setText(_('Wrong passphrase, please retry!\nEnter container passphrase:'))
            self.buttons.setEnabled(True)
            self.pw_box.setText('')
            self.pw_box.setEnabled(True)
            self.pw_box.setFocus()
            self.waiting_for_response = False

    def closeEvent(self, event):
        """ Event handler close: block while waiting for response or notify worker with abort message """
        if self.waiting_for_response:
            event.ignore()
        else:
            self.worker.execute({'type': 'abort', 'msg': ''}, None, None)

    def communicate(self):
        """ Dialog runs itself and throws an exception if the container wasn't unlocked
            :raises: UserInputError
        """
        try:
            if not self.exec_():
                raise UserInputError(self.error_message)  # empty string if cancel/delete -> won't get displayed
        finally:
            self.destroy()



"""
luckyLUKS is a GUI for creating and (un-)locking LUKS volumes from container files.
For more information visit: http://github.com/jas-per/luckyLUKS

Copyright (c) 2014,2015,2022 Jasper van Hoorn (muzius@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <http://www.gnu.org/licenses/>
"""

import locale
import pkgutil
import io
from gettext import GNUTranslations, NullTranslations

from luckyLUKS import main

# This is the entry point, when run from the zip-file package.
# To access resources inside the zip file, pkgutil.get_data() has to be used.
# Because of this, gettext will be initialized here,
# to search for a .mo file for the users locale inside the zip
if __name__ == '__main__':

    locale.setlocale(locale.LC_ALL, '')
    loc, enc = locale.getlocale(locale.LC_MESSAGES)
    l10n_resource = None

    # try to find the corresponding gettext file (*.mo) for the users locale in the zip file
    if loc is not None and loc != 'C':
        try:
            l10n_resource = pkgutil.get_data(
                'luckyLUKS',
                'locale/{0}/LC_MESSAGES/luckyLUKS.mo'.format(loc)
            )
        except IOError:
            if '_' in loc:
                try:
                    l10n_resource = pkgutil.get_data(
                        'luckyLUKS',
                        'locale/{0}/LC_MESSAGES/luckyLUKS.mo'.format(loc.split('_')[0])
                    )
                except IOError:
                    pass

    if l10n_resource is None:
        translation = NullTranslations()
    else:
        translation = GNUTranslations(io.BytesIO(l10n_resource))

    main.luckyLUKS(translation)



""" config for setuptools/stdeb """

from setuptools import setup
from luckyLUKS import VERSION_STRING

long_desc = """luckyLUKS is a Linux GUI for creating and (un-)locking
encrypted volumes from container files. Unlocked containers leave
an icon in the systray as a reminder to close them eventually ;)
Supports cryptsetup/LUKS and Truecrypt container files.

The container is basically a large file that encapsulates an
encrypted partition. This simplifies handling and backup of
encrypted data for casual users.
luckyLUKS follows a keep-it-simple philosophy for creating and using
encrypted containers, that aims to keep users from shooting themselves
in the foot. For quick access the GUI offers to add a shortcut for
unlocking a specific container to the start menu or on the desktop.

For more information and a complete FAQ see
https://github.com/jas-per/luckyLUKS"""


setup(name='luckyLUKS',
      version=VERSION_STRING,
      author='Jasper van Hoorn',
      author_email='muzius@gmail.com',
      url='https://github.com/jas-per/luckyLUKS',
      download_url='https://github.com/jas-per/luckyLUKS',
      description='GUI for encrypted LUKS or TrueCrypt containers',
      long_description=long_desc,
      platforms=['Linux'],
      packages=['luckyLUKS'],
      package_data={'luckyLUKS': ['locale/*/LC_MESSAGES/*']},
      scripts=['luckyluks'],
      keywords='python tools utils cryptsetup LUKS TrueCrypt encryption container block device mapper GUI tcplay',
      license='GPL',
      classifiers=['Development Status :: 5 - Production/Stable',
                   'Environment :: X11 Applications :: Qt',
                   'Intended Audience :: End Users/Desktop',
                   'Natural Language :: English',
                   'Natural Language :: German',
                   'Operating System :: POSIX :: Linux',
                   'Programming Language :: Python :: 3',
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7',
                   'Programming Language :: Python :: 3.8',
                   'Programming Language :: Python :: 3.9',
                   'Programming Language :: Python :: 3.10',
                   'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                   'Topic :: Utilities',
                   'Topic :: Security :: Cryptography',
                   ],
      )



"""
luckyLUKS is a GUI for creating and (un-)locking LUKS volumes from container files.
For more information visit: http://github.com/jas-per/luckyLUKS

Copyright (c) 2014,2015,2022 Jasper van Hoorn (muzius@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <http://www.gnu.org/licenses/>
"""

import os.path
import sys
import argparse
import types
import builtins
from luckyLUKS import VERSION_STRING, PROJECT_URL


def luckyLUKS(translation, *args, **kwargs):
    """ main entry point: initialize gettext, parse arguments and start either GUI or worker """

    translation.gettext_qt = types.MethodType(lambda self, msg: self.gettext(msg).replace('\n', '<br>'), translation)
    argparse._ = _ = translation.gettext  # gettext for argsparse

    parser = argparse.ArgumentParser(
        description=_('GUI for creating and unlocking LUKS/TrueCrypt volumes from container files'),
        epilog=_('When called without any arguments a setup dialog will be shown before unlocking,\n'
                 'where you can select containerfile and name, or create a new encrypted container.\n'
                 'If both arguments are supplied, the unlock dialog will be shown directly.\n\n'
                 'Example:\n'
                 '  {executable} -c /usbstick/encrypted.bin -n mydata -m /home/user/enc\n\n'
                 'If automatic mounting (eg udisks/polkit) is configured on your system,\n'
                 'explicitly setting a mountpoint is usually not needed (but still possible)\n\n'
                 'Homepage: {project_url}').format(executable=os.path.basename(sys.argv[0]),
                                                   project_url=PROJECT_URL),
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    # L10n: used by argsparse to generate help output on the console (luckyLUKS --help)
    ap_usage = _('usage: ')
    # L10n: used by argsparse to generate help output on the console (luckyLUKS --help)
    ap_optargs = _('optional arguments')
    # L10n: used by argsparse to generate help output on the console (luckyLUKS --help)
    ap_helpmsg = _('show this help message and exit')
    # L10n: used by argsparse to generate help output on the console (luckyLUKS --help)
    ap_errmsg = _('%(prog)s: error: %(message)s\n')
    # L10n: used by argsparse to generate help output on the console (luckyLUKS --help)
    ap_unknown = _('unrecognized arguments: %s')

    # if argument not specified or empty set value to None
    # error messages will be shown by the GUI, not on the command line
    parser.add_argument('-c', dest='container', nargs='?', metavar=_('PATH'),
                        help=_('Path to the encrypted container file'))
    parser.add_argument('-n', dest='name', nargs='?', metavar=_('NAME'),
                        help=_('Choose a device name to identify the unlocked container'))
    parser.add_argument('-m', dest='mountpoint', nargs='?', metavar=_('PATH'),
                        help=_('Where to mount the encrypted filesystem'))
    parser.add_argument('-k', dest='keyfile', nargs='?', metavar=_('PATH'),
                        help=_('Path to an optional key file'))
    parser.add_argument('-v', '--version', action='version', version="luckyLUKS " + VERSION_STRING,
                        help=_("show program's version number and exit"))
    parser.add_argument('--ishelperprocess', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--sudouser', type=int, help=argparse.SUPPRESS)

    parsed_args = parser.parse_args()

    # worker will be created by calling the script again (but this time with su privileges)
    builtins._ = translation.gettext_qt
    if parsed_args.ishelperprocess:
        startWorker(parsed_args.sudouser)
    else:
        startUI(parsed_args)


def startUI(parsed_args):
    """ Import the required GUI elements and create main window """
    from PyQt5.QtWidgets import QApplication
    from PyQt5.QtCore import QLocale, QTranslator, QLibraryInfo
    from luckyLUKS.mainUI import MainWindow

    # l10n qt-gui elements
    qt_translator = QTranslator()
    qt_translator.load('qt_' + QLocale.system().name(), QLibraryInfo.location(QLibraryInfo.TranslationsPath))
    application = QApplication(sys.argv)
    application.installTranslator(qt_translator)

    # start application
    main_win = MainWindow(parsed_args.name, parsed_args.container, parsed_args.keyfile, parsed_args.mountpoint)
    # setup OK -> run event loop
    if main_win.is_initialized:
        sys.exit(application.exec_())
    else:
        sys.exit(0)


def startWorker(sudouser=None):
    """ Initialize worker process """
    from luckyLUKS import worker
    if sudouser is not None:
        # helper called with su to configure sudo
        if os.getuid() == 0 and os.getenv("SUDO_UID") is None:
            try:
                worker.WorkerHelper().modify_sudoers(sudouser)
                sys.exit(0)
            except worker.WorkerException as we:
                sys.stdout.write(str(we))
                sys.exit(2)
        else:
            # deny giving other user userids sudo access to luckyLUKS if not called with su
            sys.exit(2)
    else:
        worker.run()
